import re
import typing as t
from typing import Any, Callable, Sequence

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
import networkx as nx
import sqlglot.expressions as sge
from sqlglot import Expression, exp
from sqlglot._typing import E
from sqlglot.dialects.dialect import Dialect
from sqlglot.errors import OptimizeError
from sqlglot.optimizer import build_scope, optimize, traverse_scope
from sqlglot.optimizer.annotate_types import annotate_types
from sqlglot.optimizer.canonicalize import canonicalize
from sqlglot.optimizer.eliminate_ctes import eliminate_ctes
from sqlglot.optimizer.eliminate_joins import eliminate_joins
from sqlglot.optimizer.eliminate_subqueries import eliminate_subqueries
from sqlglot.optimizer.merge_subqueries import merge_subqueries
from sqlglot.optimizer.normalize import normalize
from sqlglot.optimizer.optimize_joins import optimize_joins
from sqlglot.optimizer.pushdown_predicates import pushdown_predicates
from sqlglot.optimizer.pushdown_projections import pushdown_projections
from sqlglot.optimizer.qualify import qualify
from sqlglot.optimizer.qualify_columns import quote_identifiers
from sqlglot.optimizer.simplify import simplify
from sqlglot.optimizer.unnest_subqueries import unnest_subqueries

from vinyl.lib.errors import VinylError, VinylErrorType
from vinyl.lib.schema import VinylSchema
from vinyl.lib.table import VinylTable
from vinyl.lib.utils.graph import DAG
from vinyl.lib.utils.text import _replace_with_dict


class Catalog(dict[str, sch.Schema]):
    """A catalog of tables and their schemas."""

    typemap = {
        dt.Int8: sge.DataType.Type.TINYINT,
        dt.Int16: sge.DataType.Type.SMALLINT,
        dt.Int32: sge.DataType.Type.INT,
        dt.Int64: sge.DataType.Type.BIGINT,
        dt.Float16: sge.DataType.Type.FLOAT,  # no halffloat
        dt.Float32: sge.DataType.Type.FLOAT,
        dt.Float64: sge.DataType.Type.DOUBLE,
        dt.Decimal: sge.DataType.Type.DECIMAL,
        dt.Boolean: sge.DataType.Type.BOOLEAN,
        dt.JSON: sge.DataType.Type.JSON,
        dt.Interval: sge.DataType.Type.INTERVAL,
        dt.Timestamp: sge.DataType.Type.TIMESTAMP,
        dt.Date: sge.DataType.Type.DATE,
        dt.Binary: sge.DataType.Type.BINARY,
        dt.String: sge.DataType.Type.TEXT,
        dt.Array: sge.DataType.Type.ARRAY,
        dt.Map: sge.DataType.Type.MAP,
        dt.UUID: sge.DataType.Type.UUID,
        dt.Struct: sge.DataType.Type.STRUCT,
    }

    def to_sqlglot_dtype(self, dtype: dt.DataType) -> str:
        if dtype.is_geospatial():
            return dtype.geotype
        else:
            default = dtype.__class__.__name__.lower()
            return self.typemap.get(type(dtype), default)

    def to_sqlglot_schema(self, schema: sch.Schema) -> dict[str, str]:
        return {name: self.to_sqlglot_dtype(dtype) for name, dtype in schema.items()}

    def to_sqlglot(self):
        return {
            name: self.to_sqlglot_schema(table.schema()) for name, table in self.items()
        }


RULES = (
    # qualify,
    pushdown_projections,
    normalize,
    unnest_subqueries,
    pushdown_predicates,
    optimize_joins,
    eliminate_subqueries,
    merge_subqueries,
    eliminate_joins,
    eliminate_ctes,
    quote_identifiers,
    annotate_types,
    canonicalize,
    simplify,
)


def get_adj_node_name(n):
    if isinstance(n, VinylTable):
        # is source
        return n.get_name()
    elif hasattr(n, "_unique_name"):
        return n._unique_name
    elif hasattr(n, "_module"):
        return f"{n._module}.{n.__name__}"
    else:
        raise ValueError(f"Node {n} not recognized")


def get_parents(tpl):
    if tpl[-1].parent is None:
        return tpl
    else:
        tpl += (tpl[-1].parent,)
    return get_parents(tpl)


def get_lowest_cte(tpl):
    parents = list(get_parents(tpl))
    options = [i for i in parents if isinstance(i, sge.CTE)]
    if options == []:
        return None
    return options[0].alias


def get_sqlglot_schema(
    vinylschema: dict[str, VinylSchema],
) -> dict[str, dict[str, str]]:
    adj_schema = {}
    for k, v in vinylschema.items():
        adj_schema[k] = v.to_dict()
    return adj_schema


def validate_qualify_columns_errors(expression: E) -> list[Exception]:
    unqualified_columns = []
    errors: list[Exception] = []
    for scope in traverse_scope(expression):
        if isinstance(scope.expression, exp.Select):
            unqualified_columns.extend(scope.unqualified_columns)
            if (
                scope.external_columns
                and not scope.is_correlated_subquery
                and not scope.pivots
            ):
                for col in scope.external_columns:
                    cte = get_lowest_cte((col,))
                    errors += [
                        OptimizeError(
                            f"""Column '{col}' could not be resolved{f" in cte: '{cte}'" if cte else ' in main clause'}. Please qualify this column."""
                        )
                    ]
            for col in scope.unqualified_columns:
                cte = get_lowest_cte((col,))
                errors += [
                    OptimizeError(
                        f"""Ambiguous column '{col}' {f" in cte: '{cte}'" if cte else 'in main clause'}. Please qualify this column."""
                    )
                ]
    return errors


def qualify_sql_item(
    ast_: E, filtered_schema: dict[str, dict[str, str]], dialect: Dialect
) -> tuple[Expression, list[Exception]]:
    qualified = qualify(
        ast_,
        dialect=dialect,
        schema=filtered_schema,
        validate_qualify_columns=False,
        quote_identifiers=False,
        identify=False,
        infer_schema=True,
    )
    qualification_errors = validate_qualify_columns_errors(qualified)

    # try different qualification patterns if necessary
    if len(qualification_errors) > 0:
        qualified = qualify(
            qualified,
            dialect=dialect,
            validate_qualify_columns=False,
            quote_identifiers=False,
            identify=False,
            infer_schema=True,
        )
        qualification_errors = validate_qualify_columns_errors(qualified)
    return qualified, qualification_errors


def lineage_select(expression: E, sel: E) -> Expression:
    if isinstance(expression, exp.Select):
        select_ = t.cast(
            exp.Expression,
            expression.select(sel, append=False, copy=False),
        )
    else:
        select_ = expression.expression
    return select_


def add_sel_edge(
    g: DAG,
    tbl_name: str,
    sel: E,
    to_: str,
    ntype: list[str],
):
    from_ = (str(tbl_name) + "." + sel.alias_or_name).replace('"', "")
    g.add_edge(from_, to_, {"ntype": set(ntype)})


def find_characters_between_equal_and_parenthesis(text):
    # Regex pattern to find characters (excluding spaces) between start or '=' and '('
    # It uses a character set [^\s]* that matches any characters except whitespace characters
    pattern = r"(?:^|(?<==))([^\s]*?)(?=\()"

    # Find all matches
    matches = re.findall(pattern, text)

    return matches


def union_helper(ast, name_override: str | None = None):
    name = name_override if name_override is not None else ast.alias
    union_search = list(ast.find_all(exp.Union))
    if len(union_search) > 0:
        unions = []
        for u in union_search:
            for attr in ["left", "right"]:
                new_ast = getattr(u, attr)
                new = [(name, new_ast)]
                if (
                    new not in unions
                    and new_ast is not None
                    and not isinstance(new_ast, exp.Union)
                ):
                    unions.extend(new)
        return unions
    else:
        new_ast = ast.this if isinstance(ast.this, exp.Select) else ast
        return [(name, new_ast)]


class SQLAstNode:
    def __init__(
        self,
        id: str = "",
        ast: Expression = Expression(),
        schema: VinylSchema = VinylSchema(ibis.schema({})),
        deps: dict[str, Any] = {},
        deps_schemas: dict[str, VinylSchema] = {},
        errors: list[VinylError] = [],
        dialect: Dialect = Dialect(),
        lineage: DAG | None = None,
    ):
        self.id: str = id
        self.ast = ast
        self.schema = schema
        self.deps = deps
        self.deps_schemas = deps_schemas
        self.dialect = dialect
        self.errors = errors
        self.lineage = lineage

    def qualify(self):
        try:
            qualified = qualify(
                self.ast,
                dialect=self.dialect,
                schema=get_sqlglot_schema(self.deps_schemas),
                validate_qualify_columns=False,
                quote_identifiers=False,
                identify=False,
                infer_schema=True,
            )
            qualification_errors = validate_qualify_columns_errors(qualified)

        except Exception as e:
            qualified = self.ast
            qualification_errors = [e]

        # try different qualification patterns if necessary
        if len(qualification_errors) > 0:
            try:
                qualified = qualify(
                    qualified,
                    dialect=self.dialect,
                    validate_qualify_columns=False,
                    quote_identifiers=False,
                    identify=False,
                    infer_schema=True,
                )
                qualification_errors = validate_qualify_columns_errors(qualified)
            except Exception as e:
                qualified = self.ast
                qualification_errors = [e]
        self.ast = qualified
        self.errors += [
            VinylError(
                node_id=self.id,
                type=VinylErrorType.PARSE_ERROR,
                msg=str(e),
                dialect=self.dialect,
            )
            for e in qualification_errors
        ]
        return self

    def optimize(self, rules: Sequence[Callable[..., Any]] = RULES):
        if self.errors != []:
            return self
        self = self.qualify()
        if self.errors != []:
            return self

        try:
            self.ast = optimize(
                self.ast,
                schema=get_sqlglot_schema(self.deps_schemas),
                dialect=self.dialect,
                rules=rules,
            )
        except Exception as e:
            self.errors.append(
                VinylError(
                    self.id,
                    VinylErrorType.OPTIMIZE_ERROR
                    if isinstance(e, OptimizeError)
                    else VinylErrorType.MISCELLANEOUS_ERROR,
                    str(e),
                    self.dialect,
                    self.ast.sql(pretty=True),
                )
            )

        return self

    def lineage_sources_helper(self):
        model_name = self.id
        model_ast = self.ast.copy()
        sources = []
        for x in model_ast.find_all(exp.CTE):
            # split sources if there are unions
            sources += union_helper(x)
            x.pop()
        for x in model_ast.find_all(exp.With):
            x.pop()
        sources += union_helper(model_ast, model_name)
        # adjust table aliases and column qualifications to be trivial to make lineage parsing easier
        for i, source in enumerate(sources):
            name, cte = source
            for tbl in cte.find_all(sge.Table):
                for col in cte.find_all(sge.Column):
                    previous_sources = [] if i == 0 else sources[: (i - 1)]
                    if (
                        col.table not in self.deps
                        and col.table
                        not in [j[0] for j in previous_sources if j[0] != name]
                        and col.table is not None
                        and col.table == tbl.alias
                    ):
                        col.set("table", tbl.name)
                if tbl.alias != "" and tbl.alias != tbl.name:
                    tbl.set("alias", tbl.name)
        return sources

    def get_lineage(self, lineage_filters: list[str] = [], overwrite: bool = False):
        if self.lineage is not None and not overwrite:
            return self.lineage

        self.lineage = DAG()

        sources = self.lineage_sources_helper()

        for k, cte in sources:
            scope = build_scope(cte)
            if not scope:
                continue
            # Get information for non-select columns
            wheres = []
            havings = []
            qualifies = []
            groupbys = []
            joins = []
            if not lineage_filters or "filter" in lineage_filters:
                for li in scope.find_all(exp.Where):
                    for lj in li.find_all(exp.Column):
                        wheres += [str(lj)]

                for li in scope.find_all(exp.Having):
                    for lj in li.find_all(exp.Column):
                        havings += [str(lj)]

                for li in scope.find_all(exp.Qualify):
                    for lj in li.find_all(exp.Column):
                        qualifies += [str(lj)]
            if not lineage_filters or "group_by" in lineage_filters:
                for li in scope.find_all(exp.Group):
                    for lj in li.find_all(exp.Column):
                        groupbys += [str(lj)]
            if not lineage_filters or "join_key" in lineage_filters:
                for li in scope.find_all(exp.Join):
                    if li.args.get("on") is not None:
                        for lj in li.args.get("on").find_all(exp.Column):
                            joins += [str(lj)]

            non_select = wheres + groupbys + havings + qualifies + joins
            non_select_type = (
                ["filter" for i in wheres]
                + ["group_by" for i in groupbys]
                + ["having" for i in havings]
                + ["qualify" for i in qualifies]
                + ["join_key" for i in joins]
            )

            # get named windows
            named_windows_cols = {}
            for w in scope.find_all(exp.Window):
                if w.name == "":
                    continue
                named_windows_cols[w.name] = [
                    c.sql(dialect=self.dialect, comments=False).replace('"', "")
                    for c in list(w.find_all(exp.Column))
                ]

            # build select graph
            expression = scope.expression
            if not hasattr(expression, "selects"):
                continue

            for sel in expression.selects:
                select_helper = lineage_select(expression, sel)
                if select_helper is None or not hasattr(select_helper, "selects"):
                    continue
                selects_ = select_helper.selects
                if len(selects_) == 0:
                    continue
                select_ = selects_[
                    0
                ]  # removes everything but the pure select clause (e.g. filters)

                # get select col type
                select_parse = find_characters_between_equal_and_parenthesis(
                    str(repr(select_))
                )
                select_parse_filter = [
                    i not in ["Select", "Alias", "Column", "Identifier", "DataType"]
                    for i in select_parse
                ]
                if any(select_parse_filter):
                    select_type = "transform"
                else:
                    select_type = "as_is"

                # build select graph
                if not lineage_filters or select_type in lineage_filters:
                    for c in select_.find_all(exp.Column):
                        add_sel_edge(
                            self.lineage,
                            k,
                            sel,
                            c.sql(dialect=self.dialect, comments=False)
                            .replace('"', "")
                            .replace("`", ""),
                            [select_type],
                        )

                # handle aliased windows
                if not lineage_filters or "transform" in lineage_filters:
                    for w in select_.find_all(exp.Window):
                        if w.alias != "" and w.alias in named_windows_cols:
                            for col_name in named_windows_cols[w.alias]:
                                add_sel_edge(
                                    self.lineage,
                                    k,
                                    sel,
                                    col_name,
                                    ["transform"],
                                )

                # build non-select graph
                for i, c2 in enumerate(non_select):
                    add_sel_edge(
                        self.lineage,
                        k,
                        sel,
                        c2.replace('"', "").replace("`", ""),
                        [non_select_type[i]],
                    )
        # remove trivial nodes
        ids = [ix for ix in self.deps] + [self.id]  # sqlglot lowercases all table names
        id_replace_dict = {
            x.lower() if str(self.dialect) != "snowflake" else x.upper(): x for x in ids
        }
        nodes_to_remove = [
            x
            for x in self.lineage.node_dict
            if x.rsplit(".", 1)[0] not in id_replace_dict
        ]
        self.lineage.remove_nodes_and_reconnect(nodes_to_remove)

        # get correct case again
        self.lineage = self.lineage.relabel(
            lambda x: _replace_with_dict(x, id_replace_dict)
        )

        # add nodes for all columns in table_graph, even if they have no connections or have errors:
        all_schemas = {**self.deps_schemas, **{self.id: self.schema}}
        for node_id, schema in all_schemas.items():
            for col in schema.names:
                graph_name = f"{node_id}.{col}"
                if (
                    graph_name not in self.lineage.node_dict and "." not in col
                ):  # remove subcolumns for now
                    self.lineage.add_node(graph_name)

        return self.lineage.reverse()


class SQLProject:
    def __init__(self, nodes: list[SQLAstNode], db_graph: DAG):
        self.nodes = nodes
        self.db_graph = db_graph

    def optimize(self):
        for node in self.nodes:
            node.optimize()

        return self

    def get_lineage(self, lineage_filters: list[str] = []):
        for node in self.nodes:
            node.get_lineage(lineage_filters)

        return self

    def stitch_lineage(
        self,
        ids: list[str] | None = None,
        predecessor_depth: int | None = None,
        successor_depth: int | None = None,
    ) -> dict[str, Any]:
        col_lineages = []
        errors = []
        selected = self._select_nodes(ids, predecessor_depth, successor_depth)
        parents = self.db_graph.get_parents(selected)
        table_lineage = self.db_graph.subgraph(
            list(set(selected + parents))
        ).to_networkx()
        for id in selected:
            for node in self.nodes:
                if node.id == id:
                    lineage = node.get_lineage()

                    # add schema to table lineage nodes
                    table_lineage.nodes[id]["schema"] = {
                        name: type(coltype).__name__
                        for name, coltype in node.schema.items()
                    }
                    for tablename, schema in node.deps_schemas.items():
                        table_lineage.nodes[tablename]["schema"] = {
                            name: type(coltype).__name__
                            for name, coltype in schema.items()
                        }
                    errors.extend(node.errors)
                    if lineage is not None:
                        col_lineages.append(lineage.to_networkx())

        return {
            "errors": errors,
            "table_lineage": nx.node_link_data(table_lineage),
            "column_lineage": nx.node_link_data(nx.compose_all(col_lineages)),
        }

    def stitch_lineage_postgres(
        self,
        ids: list[str] | None = None,
        predecessor_depth: int | None = None,
        successor_depth: int | None = None,
    ):
        selected = self._select_nodes(ids, predecessor_depth, successor_depth)
        columns = []
        column_links = []
        for id in selected:
            for node in self.nodes:
                if node.id == id:
                    for name, coltype in node.schema.items():
                        col = {
                            "table_unique_name": node.id,
                            "name": name,
                            "type": str(coltype),
                        }
                        if col not in columns:
                            columns.append(col)
                    for tablename, schema in node.deps_schemas.items():
                        # if ".sources." in tablename:
                        for name, coltype in schema.items():
                            col = {
                                "table_unique_name": tablename,
                                "name": name,
                                "type": str(coltype),
                            }
                            if col not in columns:
                                columns.append(col)

                    full_lineage = node.get_lineage()
                    if not full_lineage:
                        continue

                    for source_index, target_index in full_lineage.g.edge_list():
                        source = full_lineage.node_dict.inv[source_index]
                        target = full_lineage.node_dict.inv[target_index]
                        column_links.append(
                            {
                                "source_name": source,
                                "target_name": target,
                                "lineage_type": "all",
                                "connection_types": list(
                                    full_lineage.g.get_edge_data(
                                        source_index, target_index
                                    )["ntype"]
                                ),
                            }
                        )
                    direct_lineage = node.get_lineage(
                        lineage_filters=["as-is", "transform"]
                    )
                    if not direct_lineage:
                        continue
                    for source_index, target_index in direct_lineage.g.edge_list():
                        source = direct_lineage.node_dict.inv[source_index]
                        target = direct_lineage.node_dict.inv[target_index]
                        column_links.append(
                            {
                                "source_name": source,
                                "target_name": target,
                                "lineage_type": "direct_only",
                                "connection_types": list(
                                    full_lineage.g.get_edge_data(
                                        source_index, target_index
                                    )["ntype"]
                                ),
                            }
                        )
        return columns, column_links

    def _select_nodes(self, ids, predecessor_depth, successor_depth):
        if ids is None:
            nodes = self.db_graph.topological_sort()
        else:
            nodes_to_consider = self.db_graph.get_ancestors_and_descendants(
                ids, predecessor_depth, successor_depth
            )
            nodes = self.db_graph.subgraph(list(nodes_to_consider)).topological_sort()
        nodes = [n for n in nodes if not isinstance(n, VinylTable)]
        return nodes

    def get_errors(self) -> list[VinylError]:
        errors = []
        for node in self.nodes:
            for error in node.errors:
                if error not in errors:
                    errors.append(error)
        return errors



from __future__ import annotations

import copy
import warnings
from collections.abc import Iterable
from typing import Any, Callable, Sequence, Set

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.types as ir
import ibis.selectors as s
from ibis import _

from vinyl.lib.column import VinylColumn, _demote_args
from vinyl.lib.column_methods import (
    ColumnBuilder,
    ColumnListBuilder,
    _process_cols_arg,
    column_type,
)
from vinyl.lib.constants import (
    _DIMENSION_LABEL,
    _METRIC_LABEL,
    _SCHEMA_LABEL,
    _TS_COL_NAME,
)
from vinyl.lib.enums import FillOptions, WindowType
from vinyl.lib.table import VinylTable
from vinyl.lib.table_methods import _adjust_fill_list, _join_with_removal
from vinyl.lib.temporal import _set_timezone
from vinyl.lib.utils.functions import _validate
from vinyl.lib.utils.text import _split_interval_string


class Metric:
    _tbl: VinylTable
    _name: str  # allows pydantic to still work
    _ts: Callable[..., Any]
    _agg: Callable[..., Any]
    _by: column_type
    _fill: FillOptions | Callable[..., Any]

    def __init__(
        self,
        tbl: VinylTable,
        agg: Callable[..., Any],
        ts: Callable[..., Any],
        by: column_type,
        fill=FillOptions.null,
    ):
        self._tbl = tbl
        self._agg = agg
        self._ts = ts
        self._by = by
        self._fill = fill

        # set name
        name = ColumnBuilder(self._tbl.tbl, self._agg, passthrough_deferred=True)._name
        if name is None or not isinstance(name, str):
            raise ValueError("Metric name must be a valid string")
        else:
            self._name = name

        # ensure ts is a timestamp
        self._ts = _set_timezone(self._tbl.tbl, self._ts)

        # ensure ts name is set to TS_COL_NAME constant
        ts_orig = copy.deepcopy(self._ts)
        self._ts = lambda t: ts_orig(t).name(_TS_COL_NAME)

    @property
    def _dimension_names(self) -> Set[str]:
        if self._by is None:
            return set()
        vinyl_by = ColumnListBuilder(self._tbl.tbl, self._by, passthrough_deferred=True)
        return set([n for n in vinyl_by._names if n is not None])

    def _to_dict(self) -> dict[str, Any]:
        return {
            "tbl": self._tbl,
            "name": self._name,
            "agg": self._agg,
            "ts": self._ts,
            "by": self._by,
            "fill": self._fill,
        }


def _get_dimension_names(metrics: Sequence[Metric]) -> Set[str]:
    return set().union(*[m._dimension_names for m in metrics])


class MetricStore:
    _all_metrics_dict: dict[str, Metric] = dict()
    _all_dimension_names: set[str] = set()

    _mutable: bool
    _default_tbl: VinylTable | None
    _trailing_intervals: list[ibis.interval] | None
    _metrics_dict: dict[str, Metric]
    _dimension_names: set[str]
    _derived_metrics: dict[str, ir.Deferred]

    # key will always be a vinylTable, but can't use it here due to a circular reference
    _tbl_dict: dict[VinylTable, dict[str, Any]] = {}

    __dict__: dict[str, Any]

    def __init__(
        self,
        metrics: Sequence[Metric] = [],
        derived_metrics: dict[str, ir.Deferred] = {},
        default_tbl: VinylTable | None = None,
    ):
        self._mutable = False
        self._default_tbl = default_tbl
        self._trailing_intervals = None
        self._metrics_dict = {m._name: m for m in metrics}
        self._derived_metrics = derived_metrics
        self._dimension_names = _get_dimension_names(metrics)

        self._tbl_dict = {}
        for m in metrics:
            if m._tbl not in self._tbl_dict:
                self._tbl_dict[m._tbl] = {}
                self._tbl_dict[m._tbl][_METRIC_LABEL] = set()
                self._tbl_dict[m._tbl][_DIMENSION_LABEL] = set()
                self._tbl_dict[m._tbl][_SCHEMA_LABEL] = {}

            self._tbl_dict[m._tbl][_METRIC_LABEL].add(m._name)
            self._tbl_dict[m._tbl][_DIMENSION_LABEL].update(m._dimension_names)
            col_helper_it = ColumnBuilder(m._tbl.tbl, m._agg, passthrough_deferred=True)
            self._tbl_dict[m._tbl][_SCHEMA_LABEL].setdefault(
                col_helper_it._name, col_helper_it._type
            )
            self._tbl_dict[m._tbl][_SCHEMA_LABEL][_TS_COL_NAME] = dt.Timestamp()
            if m._by is not None and isinstance(m._by, Iterable):
                for dim in ColumnListBuilder(
                    m._tbl.tbl, m._by, passthrough_deferred=True
                ):
                    self._tbl_dict[m._tbl][_SCHEMA_LABEL].setdefault(
                        dim._name, dim._type
                    )

        MetricStore._address_naming_conflicts(self._metrics_dict)
        MetricStore._append(self._metrics_dict)

        self._set_columns()

    # create column function autocomplete
    def __getattr__(self, name) -> VinylColumn:
        return self.__getattribute__(name)

    def _copy(self) -> MetricStore:
        return MetricStore(
            list(self._metrics_dict.values()), self._derived_metrics, self._default_tbl
        )

    def _set_columns(self):
        # set columns
        count = 0
        for name in ["ts", *self._dimension_names, *self._metrics_dict]:
            col = VinylColumn(getattr(_, name))
            setattr(self, name, col)
            setattr(self, f"_{count}", col)
            count += 1

        for k, v in self._derived_metrics.items():
            setattr(self, k, v)
            setattr(self, f"_{count}", v)
            count += 1

    def _clear_columns(self):
        for name in self.__dict__.copy():
            if name not in [
                "_mutable",
                "_default_tbl",
                "_trailing_intervals",
                "_metrics_dict",
                "_dimension_names",
                "_tbl_dict",
                "_derived_metrics",
            ]:
                delattr(self, name)

    def _reset_columns(self):
        self._clear_columns()
        self._set_columns()

    def __add__(self, other: MetricStore) -> MetricStore:
        if not isinstance(other, MetricStore):
            raise ValueError(
                f"Can only add two MetricStores together, not a {type(other)}"
            )
        if self._default_tbl is None:
            return other
        if other._default_tbl is None:
            return self

        if self._default_tbl != other._default_tbl:
            raise ValueError(
                "Can't add two MetricStores with different default tables. Use a join (e.g. tbl1 * tbl2) instead."
            )
        combined_metrics = list(self._metrics_dict.values()) + list(
            other._metrics_dict.values()
        )
        combined_derived_metrics = {**self._derived_metrics, **other._derived_metrics}

        out = MetricStore(
            combined_metrics,
            combined_derived_metrics,
            self._default_tbl,
        )
        out._reset_columns()
        if self._mutable:
            self.__dict__ = out.__dict__
            self.__dict__["_mutable"] = True  # reset mutability to True
        return out

    def __radd__(self, other: MetricStore) -> MetricStore:
        if not isinstance(other, MetricStore):
            raise ValueError(
                f"Can only add two MetricStores together, not a {type(other)}"
            )
        return other.__add__(self)

    # Make callable so that @model wrapper works
    def __call__(self):
        return self

    # Make mutable when using in a context manager
    def __enter__(self):
        # Create a copy of the original object and make the object mutable
        new = MetricStore(list(self._metrics_dict.values()), self._default_tbl)
        new._mutable = True
        return new

    def __exit__(self, exc_type, exc_value, traceback):
        # Exit logic here
        pass

    def select(
        self,
        cols: column_type,
        trailing: list[int | None] = [None],
    ) -> VinylTable:
        """
        Retrieves the selected columns from the MetricStore.

        For `cols`, you may select metrics, dimensions, or the timestamp column. You can also select any expressions (e.g. met1 + met2) based on metrics or dimensions, but not a mix of metrics and dimensions in the same expression. If you select the timestamp column, you must use the `.floor` method to bucket the timestamps to a desired interval.

        If no trailing intervals are provided, the function will return the aggregated value for each col. If you provide trailing intervals, the function will return the aggregated value over the trailing intervals. For example, if you provide [1, 2], the function will return the aggregated value **for each `col`** over the interval and the previous interval, and the aggregated value over the interval and the previous two intervals.
        """
        if self._default_tbl is None:
            raise ValueError("Can't select from an empty MetricStore.")
        cols_adj = _process_cols_arg(
            self._default_tbl.tbl, cols, passthrough_deferred=True
        )
        return MetricSelect(self, cols=cols_adj, intervals=trailing)._select()

    def metric(
        self,
        cols: ir.Scalar
        | Callable[..., ir.Scalar]
        | list[ir.Scalar | Callable[..., ir.Scalar]]
        | dict[str, ir.Scalar | Callable[..., ir.Scalar]],
        ts: ir.TimestampValue | Callable[..., ir.TimestampValue],
        by: ir.Value
        | Callable[..., ir.Value]
        | Sequence[ir.Value | Callable[..., ir.Value]] = [],
        fill: FillOptions
        | Callable[..., Any] = FillOptions.null,  # all metrics have fill
        tbl: VinylTable | None = None,
    ):
        if tbl is None:
            if self._default_tbl is None:
                raise ValueError(
                    "No default table set. Please provide a table to the metric function."
                )
            tbl = self._default_tbl
        vinyl_cols = ColumnListBuilder(tbl.tbl, cols, passthrough_deferred=True)
        if len(vinyl_cols._cols) == 0:
            raise ValueError("Must provide at least one metric to metric function")

        vinyl_by = ColumnListBuilder(tbl.tbl, by, passthrough_deferred=True)
        vinyl_ts = ColumnBuilder(tbl.tbl, ts, passthrough_deferred=True)

        fill_list = _adjust_fill_list(len(vinyl_cols._cols), fill)

        mets = []
        for i, v in enumerate(vinyl_cols):
            met = Metric(
                tbl=tbl,
                agg=v._lambdaized,
                ts=vinyl_ts._lambdaized,
                by=vinyl_by._lambdaized,
                fill=fill_list[i],
            )
            mets.append(met)
        out = MetricStore(metrics=mets, derived_metrics={}, default_tbl=tbl)
        combined = self + out
        return combined

    def derive(
        self,
        metrics: ir.Scalar
        | Callable[..., ir.Scalar]
        | list[ir.Scalar | Callable[..., ir.Scalar]]
        | dict[str, ir.Scalar | Callable[..., ir.Scalar]],
    ):
        """
        Add a derived metric to the MetricStore. Derived metrics are calculated from existing metrics in the MetricStore.

        Please do not include any other columns in the derived metric. This could cause unexpected behavior.
        """
        # build mock tbl to help with getting column names
        mock_cols = []
        for name, met in self._metrics_dict.items():
            type_ = ColumnBuilder(met._tbl.tbl, met._agg)._type
            mock_cols.append((name, type_))
        mock_tbl = ibis.table(ibis.schema(mock_cols))
        x = ColumnListBuilder(mock_tbl, metrics, passthrough_deferred=True)

        # build derived metrics dict
        new_derived_metrics = {
            col._name: col._col
            for col in x._cols
            if col._name is not None and col._col is not None
        }
        for k in new_derived_metrics:
            if k in self._metrics_dict:
                raise ValueError(
                    f"Derived metric name {k} already exists in the global metric store. Please rename the derived metric."
                )

        # union with self
        out = self + MetricStore(
            derived_metrics=new_derived_metrics, default_tbl=self._default_tbl
        )
        return out

    def _trailing(self, intervals: list[int | None] = [None]):
        if isinstance(intervals[0], str):
            intervals = [_split_interval_string(i) for i in intervals]

        ## arguments here should be like (1, "d")
        self._trailing_invervals = [ibis.interval(*i) for i in intervals]

    @classmethod
    def _append(cls, new_metrics_dict):
        cls._all_metrics_dict.update(new_metrics_dict)
        cls._all_dimension_names = cls._all_dimension_names | _get_dimension_names(
            new_metrics_dict.values()
        )

    @classmethod
    def _address_naming_conflicts(cls, new_metrics_dict: dict[str, Metric]):
        # set warnings for metric name conflicts
        all_intersections = set(new_metrics_dict.keys()) & set(
            cls._all_metrics_dict.keys()
        )
        nontrivial_intersections = set()
        for i in all_intersections:
            new_metrics_entry = new_metrics_dict[i]
            cls_metrics_entry = cls._all_metrics_dict[i]
            if hash(new_metrics_entry._agg(new_metrics_entry._tbl.tbl)) != hash(
                cls_metrics_entry._agg(cls_metrics_entry._tbl.tbl)
            ):
                nontrivial_intersections.add(i)

        if any(nontrivial_intersections):
            warning_txt = f"\n\nMetric(s) {nontrivial_intersections} already exist in the global metric store. The new metric(s) will overwrite the old metric(s).\n\n Here's how the formulas will change:\n"
            for k, v in new_metrics_dict.items():
                if k in nontrivial_intersections:
                    warning_txt += f"- {k}: {cls._all_metrics_dict[k]._agg(cls._all_metrics_dict[k]._tbl.tbl)} -> {v._agg(v._tbl.tbl)}\n"
            warnings.warn(message=warning_txt)

        # raise error when metric name conflicts with dimension name
        new_all_metric_names = set(new_metrics_dict.keys()) | set(
            cls._all_metrics_dict.keys()
        )
        new_all_dimension_names = _get_dimension_names(
            [v for v in new_metrics_dict.values()]
        ) | set(cls._all_dimension_names)

        if any(new_all_metric_names & new_all_dimension_names):
            raise ValueError(
                f"\n\nMetric name(s) {new_all_metric_names} conflict with dimension name(s) {new_all_dimension_names}. Please rename the metric(s) or dimension(s)."
            )


class MetricSelect:
    _metric_store: MetricStore
    _intervals: list[int | None]
    _ts: list[ir.Value]
    _dimensions: list[ir.Value]
    _temp_dimensions: list[ir.Value]
    _temp_metrics: list[ir.Value]
    _tbl_ts_metrics_to_select: dict[VinylTable, list[ir.Value]]

    _raw_metric_tbls: list[VinylTable]
    _all_exprs: list[ir.Value]
    _combined_schema: VinylTable

    def __init__(
        self,
        MetricStore: MetricStore,
        cols: list[ir.Value],
        intervals: list[int | None] = [None],
    ):
        self._metric_store = MetricStore
        self._intervals = intervals
        self._ts = []
        self._dimensions = []
        self._temp_dimensions = []
        self._temp_metrics = []

        # helper to select metrics from the same tbl together as long as they have the same ts
        self._tbl_ts_metrics_to_select = {}

        # helper to store raw metric tables
        self._raw_metric_tbls = []

        self._all_exprs = cols

        # create combined schema
        for i, vals in enumerate(list(self._metric_store._tbl_dict.values())):
            if i == 0:
                self._combined_schema = VinylTable._create_from_schema(
                    vals[_SCHEMA_LABEL]
                )
            else:
                self._combined_schema += VinylTable._create_from_schema(
                    vals[_SCHEMA_LABEL]
                )

    def _process_unaltered_col(self, col, col_name):
        vinyl_col = ColumnBuilder(
            self._combined_schema.tbl, col, passthrough_deferred=True
        )
        if col_name in self._metric_store._dimension_names:
            self._dimensions.append(vinyl_col._lambdaized)
        elif col_name in self._metric_store._metrics_dict:
            met = self._metric_store._metrics_dict[col_name]
            key_it = met._tbl.define(met._ts(met._tbl))  # use ts column as unique
            self._tbl_ts_metrics_to_select.setdefault(key_it, set())
            self._tbl_ts_metrics_to_select[key_it].add(col_name)
        elif col_name == _TS_COL_NAME:
            raise ValueError(
                "Can't select the raw ts column. Consider using the .truncate() or .bucket() method"
            )
        else:
            raise ValueError(f"{col_name} is not a valid column name")

    def _process_derived_col(self, col, col_name):
        vinyl_col = ColumnBuilder(
            self._combined_schema.tbl, col, passthrough_deferred=True
        )
        sources = vinyl_col._sources
        source_types = []
        for source in sources:
            if source in self._metric_store._dimension_names:
                source_types.append(_DIMENSION_LABEL)
            elif source in self._metric_store._metrics_dict:
                source_types.append(_METRIC_LABEL)
            elif source == _TS_COL_NAME:
                source_types.append(_TS_COL_NAME)
            else:
                raise ValueError(f"{source} is not a valid column name")

            if len(set(source_types)) > 1:
                raise ValueError(
                    "Can't mix metrics and dimensions in the same expression"
                )
            elif source_types[0] == _TS_COL_NAME:
                self._ts.append(vinyl_col._lambdaized)
            elif source_types[0] == _DIMENSION_LABEL:
                self._temp_dimensions.append(vinyl_col._lambdaized)
                # ensure that sources are added to dimensions so derivation works later
                self._dimensions.append(source)
            elif source_types[0] == _METRIC_LABEL:
                self._temp_metrics.append(vinyl_col._lambdaized)
                # ensure that sources are added to metrics so derivation works later
                met_source = self._metric_store._metrics_dict[source]
                key_it = met_source._tbl.define(
                    met_source._ts(met_source._tbl)
                )  # use ts column as unique so that the same tables with different ts columns are aggregated separately
                self._tbl_ts_metrics_to_select.setdefault(key_it, set())
                self._tbl_ts_metrics_to_select[key_it].add(source)
            else:
                raise ValueError(f"{col_name} is not a valid column name")

    def _process_col(self, col):
        vinyl_col = ColumnBuilder(
            self._combined_schema.tbl, col, passthrough_deferred=True
        )
        col_name = vinyl_col._name
        mets_to_select = []
        for i in self._tbl_ts_metrics_to_select.values():
            mets_to_select.extend(list(i))
        cur_col_names_helper = [
            d
            for d in list(self._dimensions)
            + list(self._temp_dimensions)
            + list(self._temp_metrics)
            + mets_to_select
        ]
        cur_col_names = ColumnListBuilder(
            self._combined_schema.tbl, cur_col_names_helper, passthrough_deferred=True
        )._names

        # skip if column is already in dimensions or metrics already selected (including derived)
        if col_name in cur_col_names:
            return

        if vinyl_col._is_unaltered:
            self._process_unaltered_col(vinyl_col._lambdaized, col_name)

        else:
            self._process_derived_col(vinyl_col._lambdaized, col_name)

    def _build_final_tbls(self):
        for interval in self._intervals:
            raw_metric_tbl_interval = []
            mets_rename_dict = {}
            # prepare raw metric tables
            for tbl, met_names in self._tbl_ts_metrics_to_select.items():
                tbl = tbl._copy()  # copy tbl to ensure no mutability issues
                mets_it = [
                    self._metric_store._metrics_dict[name]._agg for name in met_names
                ]
                fill_it = [
                    self._metric_store._metrics_dict[name]._fill for name in met_names
                ]
                vinyl_cols = ColumnListBuilder(
                    tbl.tbl, mets_it, passthrough_deferred=True
                )
                source_cols = vinyl_cols._sources_as_deferred_resolved(tbl.tbl)
                vinyl_dims = ColumnListBuilder(
                    tbl.tbl, self._dimensions, passthrough_deferred=True
                )
                source_dims = vinyl_dims._names_as_deferred
                vinyl_ts = ColumnListBuilder(
                    tbl.tbl, self._ts, passthrough_deferred=True
                )
                source_ts = vinyl_ts._names
                ts_original_source = vinyl_ts._get_direct_col_sources(unique=True)

                if interval is None:
                    tbl = tbl.select(
                        source_cols,
                        by=[col(tbl.tbl) for col in self._dimensions],
                        sort=[col(tbl.tbl) for col in self._ts],
                    )

                    out_it = tbl.aggregate(
                        cols=[met(tbl.tbl) for met in mets_it],
                        by=source_dims,
                        sort=source_ts,
                        fill=fill_it,
                    )
                else:
                    tbl = tbl.aggregate_all(
                        col_selector=source_cols,
                        f=lambda x: x.collect(),
                        by=[col(tbl.tbl) for col in self._dimensions],
                        sort=[col(tbl.tbl) for col in self._ts],
                        fill=FillOptions.null,  # trailing metrics only support null fill,
                    )
                    collected_tbl_with_trailing = tbl.select_all(
                        col_selector=vinyl_cols._sources_as_deferred_resolved(tbl.tbl),
                        f=lambda x: x.collect().array.flatten(),
                        by=[dim.resolve(tbl.tbl) for dim in source_dims],
                        sort=vinyl_ts._names_as_deferred_resolved(tbl.tbl),
                        window_type=WindowType.rows,
                        window_bounds=(-interval, 0),
                        rename=False,
                    )

                    ## NOTE: to replace once Vinyl mutate function is written
                    unnested = collected_tbl_with_trailing.define_all(
                        col_selector=[s.of_type(dt.Array)],
                        f=[lambda y: y.array.unnest()],
                        rename=False,
                    ).rename(
                        {v: source_ts[i] for i, v in enumerate(ts_original_source)}
                    )  # rename ts to make sure transformation is available for the fill in next step

                    out_it = unnested.aggregate(
                        cols=[met(unnested.tbl) for met in mets_it],
                        by=[dim.resolve(unnested.tbl) for dim in source_dims],
                        sort=[col(unnested.tbl) for col in self._ts],
                        fill=FillOptions.null,
                    )

                    mets_rename_dict.update()

                raw_metric_tbl_interval.append(out_it)

            # join together
            joined_raw_metric_tbl_interval = raw_metric_tbl_interval[0]
            for tbl in raw_metric_tbl_interval[1:]:
                # Takes connection replace of first table, this will fail if there are two different connections in the list
                joined_raw_metric_tbl_interval = VinylTable(
                    _join_with_removal(
                        joined_raw_metric_tbl_interval.tbl, tbl.tbl
                    )._arg,
                    _conn_replace=joined_raw_metric_tbl_interval._conn_replace,
                    _twin_conn_replace=joined_raw_metric_tbl_interval._twin_conn_replace,
                    _col_replace=joined_raw_metric_tbl_interval._col_replace,
                )

            # calculate derived dimensions and metrics
            for col in self._temp_dimensions + self._temp_metrics:
                joined_raw_metric_tbl_interval = joined_raw_metric_tbl_interval.define(
                    col(joined_raw_metric_tbl_interval.tbl)
                )

            # reorder to match original requested order
            cols_to_select = [
                col.resolve(joined_raw_metric_tbl_interval.tbl)
                for col in ColumnListBuilder(
                    self._combined_schema.tbl,
                    self._all_exprs,
                    passthrough_deferred=True,
                )._names_as_deferred
            ]
            joined_raw_metric_tbl_interval = joined_raw_metric_tbl_interval.select(
                cols_to_select
            )

            # rename trailing metrics
            if interval is not None:
                to_rename = [
                    i.get_name()
                    for i in s.where(
                        lambda x: x.get_name() not in source_ts + vinyl_dims._names
                    ).expand(joined_raw_metric_tbl_interval.tbl)
                ]
                joined_raw_metric_tbl_interval = joined_raw_metric_tbl_interval.rename(
                    {f"{v}_{interval}": v for v in to_rename}
                )
            self._raw_metric_tbls.append(joined_raw_metric_tbl_interval)

    def _select(self):
        for col in self._all_exprs:
            self._process_col(col)

        # build raw metric tables
        self._build_final_tbls()

        # join raw metric tables
        joined_raw_metric_tbl = self._raw_metric_tbls[0]
        for tbl in self._raw_metric_tbls[1:]:
            # Takes connection replace of first table, this will fail if there are two different connections in the list
            joined_raw_metric_tbl = VinylTable(
                _join_with_removal(joined_raw_metric_tbl.tbl, tbl.tbl)._arg,
                _conn_replace=joined_raw_metric_tbl._conn_replace,
                _twin_conn_replace=joined_raw_metric_tbl._twin_conn_replace,
                _col_replace=joined_raw_metric_tbl._col_replace,
            )

        # sort outputs
        final = joined_raw_metric_tbl.sort(s.all())

        return final


# Add pydantic validations. Need to do after function definition to make sure return class (VinylTable) is available to pydantic. Need to demote_args outside of the function to ensure validate works correctly
MetricStore.select = _demote_args(_validate(MetricStore.select))  # type: ignore[method-assign]
MetricStore.metric = _demote_args(_validate(MetricStore.metric))  # type: ignore[method-assign]



import inspect
import json
import os
from typing import Any

import ibis.expr.datatypes
import typer
from rich.console import Console
from rich.table import Table
from tqdm import tqdm

from vinyl.cli.events import Event, EventLogger
from vinyl.lib.connect import (
    DatabaseFileConnector,
    SourceInfo,
    _DatabaseConnector,
    _FileConnector,
    _TableConnector,
)
from vinyl.lib.definitions import _load_project_defs
from vinyl.lib.project import Project
from vinyl.lib.settings import _get_project_module_name, _load_project_module
from vinyl.lib.source import _get_twin_relative_path
from vinyl.lib.utils.ast import (
    _find_classes_and_attributes,
    _get_imports_from_file_regex,
)
from vinyl.lib.utils.files import _create_dirs_with_init_py, ruff_format_text
from vinyl.lib.utils.functions import _with_modified_env
from vinyl.lib.utils.obj import (
    is_valid_class_name,
    table_to_python_class,
    to_valid_class_name,
)
from vinyl.lib.utils.text import _make_python_identifier, _replace_with_dict

console = Console()


sources_cli = typer.Typer(pretty_exceptions_show_locals=False)


@sources_cli.command("list")
def list_sources(tables: bool = False):
    """Caches sources to a local directory (default: .turntable/sources)"""
    defs = _load_project_defs()
    project = Project(resources=defs.resources, models=defs.models)

    table = Table("Name", "Resource", "Location", title="Sources")
    for source in project._get_source_objects()[0]:
        if source._parent_resource is None:
            raise ValueError("Source must have a parent resource.")
        table.add_row(
            f"[bold]{source._name}[bold]",
            f"[grey70]{source._parent_resource.def_.__name__}[grey70]",
            f"[grey70]{source._location}[grey70]",
        )
    console.print(table)


def get_type_class_replace_dict():
    types = ibis.expr.datatypes
    to_replace = [k for k, v in types.__dict__.items() if inspect.isclass(v)]
    return {f"{k}(": f"t.{k}(" for k in to_replace}


def source_to_class_string(
    source: SourceInfo,
    saved_attributes: dict[str, str],
    generate_twin: bool = False,
    root_path: str | None = None,
    sample_size: int = 1000,
    twin: bool = False,
) -> str:
    if not root_path:
        module_file = _load_project_module().__file__
        if module_file is None:
            raise ValueError("Unable to find the module path for the current project")

    class_name = table_to_python_class(source._name)
    class_body = f'    _table = "{source._name}"\n'
    pr = source._parent_resource
    proj_name = _get_project_module_name()

    if pr is None:
        raise ValueError("Source must have a parent resource.")

    if isinstance(pr.connector, _TableConnector):
        if twin:
            tbl = pr.connector._get_table(source._location)
        class_body += (
            f'    _unique_name = "{proj_name}.sources.{pr.name}.{class_name}"\n'
        )

    elif isinstance(pr.connector, _DatabaseConnector):
        # source is a database
        database, schema = source._location.split(".")
        class_body += f'    _unique_name = "{proj_name}.sources.{pr.name}.{_make_python_identifier(database)}.{_make_python_identifier(schema)}.{class_name}"\n'
        class_body += f'    _schema = "{schema}"\n'
        class_body += f'    _database = "{database}"\n'

        database, schema = source._location.split(".")
        if twin:
            tbl = pr.connector._get_table(database, schema, source._name)
        class_body += (
            f'    _twin_path = "{_get_twin_relative_path(pr.name, database, schema)}"\n'
        )
        # need row count if using local files (since sampling is done live from the file)

    else:
        raise NotImplementedError(
            f"Connector type {type(pr.connector)} is not yet supported"
        )

    # need row count if using local files (since sampling is done live from the file)
    if isinstance(pr.connector, _FileConnector):
        if isinstance(pr.connector, DatabaseFileConnector):
            # in this case, location is not the real path, but the database and schema, but we can use the path from the connector
            path = os.path.relpath(pr.connector._path, root_path)
        else:
            path = os.path.relpath(source._location, root_path)
        class_body += f'    _path = "{path}"\n'
        if not pr.connector._remote and twin:
            try:
                count = tbl.count().execute()
                # in this case, we will not be caching the table, so we need the row_count
            except Exception:
                try:
                    # try to count the first column
                    first_col = tbl.columns[0]
                    count = tbl.aggregate(tbl[first_col].count()).execute().iloc[0, 0]
                except Exception:
                    # if we can't get the row count, we will just use a small default value, so we get a full sample
                    count = 0.01
            class_body += f"    _row_count = {count}\n"

    if source._schema is None:
        raise ValueError(f"Schema for {source._name} is not available")

    col_body = ""
    prev_replacements: list[str] = []
    col_replace_dict = {}
    types_replace_dict = get_type_class_replace_dict()
    for col_name, col_type in source._schema.items():
        if not is_valid_class_name(col_name):
            old_col_name = col_name
            col_name = to_valid_class_name(old_col_name, prev_replacements)
            prev_replacements.append(col_name)
            col_replace_dict[col_name] = old_col_name

        # get the string representation of the type, replacing each standard type with t.type
        col_type_str = str(col_type.__repr__())
        col_type_str = _replace_with_dict(col_type_str, types_replace_dict)
        col_type_str = _replace_with_dict(
            col_type_str,
            {"'t.Struct": "t.Struct", "t.Struct(": "t.Struct.from_tuples("},
        )  # special case for Struct, fixes bug in repr with extra quote

        base = f"    {col_name}: {col_type_str}"
        if col_name in saved_attributes:
            base += f" = {saved_attributes[col_name]}"
        col_body += f"{base}\n"

    class_body += f"    _col_replace = {json.dumps(col_replace_dict)}" + "\n\n"
    class_body += col_body

    out = f"""class {class_name}:
{class_body}
"""
    return out


def source_to_class_dict(
    source: SourceInfo,
    saved_attributes: dict[str, str],
    generate_twin: bool = False,
    root_path: str | None = None,
    sample_size: int = 1000,
    twin: bool = False,
) -> dict[str, Any]:
    if not root_path:
        module_file = _load_project_module().__file__
        if module_file is None:
            raise ValueError("Unable to find the module path for the current project")

    out: dict[str, Any] = {}
    out["config"] = {}
    out["name"] = table_to_python_class(source._name)
    out["config"]["table"] = source._name
    out["resource_name"] = source._parent_resource.name
    pr = source._parent_resource

    proj_name = _get_project_module_name()

    if pr is None:
        raise ValueError("Source must have a parent resource.")

    if isinstance(pr.connector, _TableConnector):
        if twin:
            tbl = pr.connector._get_table(source._location)
        out["unique_name"] = f"{proj_name}.sources.{pr.name}.{out['name']}"

    elif isinstance(pr.connector, _DatabaseConnector):
        # source is a database
        database, schema = source._location.split(".")
        out["unique_name"] = (
            f"{proj_name}.sources.{pr.name}.{_make_python_identifier(database)}.{_make_python_identifier(schema)}.{out['name']}"
        )
        out["read_only"] = True
        out["type"] = "source"
        out["config"]["table_schema"] = schema
        out["config"]["database"] = database

        database, schema = source._location.split(".")
        if twin:
            tbl = pr.connector._get_table(database, schema, source._name)
        out["config"]["twin_path"] = _get_twin_relative_path(pr.name, database, schema)
        # need row count if using local files (since sampling is done live from the file)

    else:
        raise NotImplementedError(
            f"Connector type {type(pr.connector)} is not yet supported"
        )

    # need row count if using local files (since sampling is done live from the file)
    if isinstance(pr.connector, _FileConnector):
        if isinstance(pr.connector, DatabaseFileConnector):
            # in this case, location is not the real path, but the database and schema, but we can use the path from the connector
            path = os.path.relpath(pr.connector._path, root_path)
        else:
            path = os.path.relpath(source._location, root_path)
        out["config"]["path"] = path
        if not pr.connector._remote and twin:
            try:
                count = tbl.count().execute()
                # in this case, we will not be caching the table, so we need the row_count
            except Exception:
                try:
                    # try to count the first column
                    first_col = tbl.columns[0]
                    count = tbl.aggregate(tbl[first_col].count()).execute().iloc[0, 0]
                except Exception:
                    # if we can't get the row count, we will just use a small default value, so we get a full sample
                    count = 0.01
            out["config"]["row_count"] = count

    if source._schema is None:
        raise ValueError(f"Schema for {source._name} is not available")

    out["config"]["columns"] = {}
    prev_replacements: list[str] = []
    col_replace_dict = {}
    for col_name, col_type in source._schema.items():
        if not is_valid_class_name(col_name):
            old_col_name = col_name
            col_name = to_valid_class_name(old_col_name, prev_replacements)
            prev_replacements.append(col_name)
            col_replace_dict[col_name] = old_col_name
        if col_name in saved_attributes:
            out["config"]["columns"][col_name] = saved_attributes[col_name]
        else:
            out["config"]["columns"][col_name] = {}

        out["config"]["columns"][col_name] = {"type": str(col_type)}

    out["config"]["col_replace"] = col_replace_dict
    return out


def _get_save_dir(sources_path: str, source: SourceInfo) -> str:
    if source._parent_resource is None:
        raise ValueError("Source must have a parent resource.")
    if isinstance(source._parent_resource.connector, _TableConnector):
        # source is a file
        return os.path.join(sources_path, source._parent_resource.name)
    if isinstance(source._parent_resource.connector, _DatabaseConnector):
        # source is a database
        identifers = [
            _make_python_identifier(str_) for str_ in source._location.split(".")
        ]
        return os.path.join(sources_path, source._parent_resource.name, *identifers)
    raise NotImplementedError(
        f"Connector type {type(source._parent_resource.connector)} is not yet supported"
    )


def generate_sources_helper(
    resources=[],
    additional_saved_attributes: dict[str, dict[str, Any]] = {},
    json: bool = False,
    defs=None,
    twin: bool = False,
) -> tuple[list[SourceInfo], list[dict[str, Any]]]:
    if defs is None:
        defs = _load_project_defs()
    if len(resources) == 0:
        project = Project(resources=defs.resources)
    else:
        project = Project(
            resources=[r for r in defs.resources if r.__name__ in resources]
        )

    sources, errors = project._get_source_objects(with_schema=True)

    if not json:
        root_file = _load_project_module().__file__
        if root_file is None:
            raise ValueError("Unable to find the module path for the current project")
        root_path = os.path.dirname(root_file)
        sources_path = os.path.join(root_path, "sources")

        _create_dirs_with_init_py(sources_path)

    source_list = []

    for source in tqdm(
        sources,
        desc="generating source files" if not json else "generating source dicts",
    ):
        if not json:
            save_dir = _get_save_dir(sources_path, source)
            file_path = os.path.join(save_dir, f"{source._name}.py")
            _create_dirs_with_init_py(save_dir)

            try:
                saved_attributes = _find_classes_and_attributes(file_path)
            except (IndexError, FileNotFoundError):
                saved_attributes = {}

            try:
                saved_attributes.update(
                    additional_saved_attributes[source._parent_resource.name][
                        table_to_python_class(source._name)
                    ]
                )
            except KeyError:
                pass

        if json:
            source_data = source_to_class_dict(
                source, {}, root_path="tenant", twin=twin
            )
            source_list.append(source_data)
            continue

        saved_imports = (
            _get_imports_from_file_regex(file_path)
            if saved_attributes != {} and os.path.exists(file_path)
            else None
        )
        with open(os.path.join(save_dir, f"{source._name}.py"), "w+") as f:
            out = ""
            if saved_imports:
                out += saved_imports
            else:
                out += "# type: ignore\n"  # prevents pylance errors tied to Ibis
                out += "from vinyl import Field, source # noqa F401\n"
                out += "from vinyl import types as t # noqa F401\n"
                out += "from ibis.common.collections import FrozenDict # noqa F401\n\n"
                out += f"from {source._parent_resource.def_.__module__} import {source._parent_resource.def_.__name__} # noqa F401 \n\n\n"
            out += f"@source(resource={source._parent_resource.name})\n"
            out += source_to_class_string(
                source, saved_attributes, root_path=root_path, twin=twin
            )
            try:
                out = ruff_format_text(out)
            except Exception as e:
                print(f"Error formatting {source._name}: {e}")
            f.write(out)
    return sources, source_list, errors


@sources_cli.command("generate")
def generate_sources(
    twin: bool = typer.Option(
        False, "--generate_twin", "-t", help="exported name of the model"
    ),
    resources: list[str] = typer.Option(
        [], "--resource", "-r", help="resource names to select"
    ),
    json: bool = typer.Option(
        False, "--json", "-j", help="output as json instead of in directory"
    ),
):
    """Generates schema files for sources"""
    track = EventLogger()
    track.log_event(Event.SOURCE_GEN)
    root_file = _load_project_module().__file__
    if root_file is None:
        raise ValueError("Unable to find the module path for the current project")
    root_path = os.path.dirname(root_file)
    sources_path = os.path.join(root_path, "sources")

    @_with_modified_env("NO_MODELS_VINYL", "True")
    def run_fn():
        sources, _, _ = generate_sources_helper(resources, json=json, twin=twin)

        if twin:
            msg = "generating twins... " if len(sources) > 1 else "generating twin... "
            for source in tqdm(
                sources,
                msg,
                unit="source",
            ):
                pr = source._parent_resource
                if isinstance(pr.connector, _DatabaseConnector):
                    database, schema = source._location.split(".")
                    pr.connector._generate_twin(
                        os.path.join(root_path, _get_twin_relative_path(pr.name)),
                        database,
                        schema,
                        source.name,
                    )
                elif isinstance(pr.connector, _TableConnector):
                    # doesn't actually generate a file, just returns the path
                    pr.connector._generate_twin(source.location)

        print(f"Generated {len(sources)} sources at {sources_path}")

    run_fn()



from __future__ import annotations

import builtins  # noqa: F401

import ibis
import ibis.expr.datatypes as types  # noqa: F401
import pydantic
import sqlglot
from ibis.expr.datatypes import *  # noqa: F401
from rich.traceback import install

from vinyl.lib.asset import (  # noqa: F401
    metric,
    model,
    resource,  # noqa: F401
)

# from vinyl.lib.definitions import Defs  # noqa: F401
from vinyl.lib.enums import FillOptions  # noqa: F401,
from vinyl.lib.field import Field  # noqa: F401
from vinyl.lib.metric import MetricStore  # noqa: F401
from vinyl.lib.set import difference, intersect, join, union  # noqa: F401
from vinyl.lib.source import source  # noqa: F401
from vinyl.lib.table import (
    VinylTable,  # noqa: F401
)
from vinyl.lib.utils.graphics import rich_print  # noqa: F401

from vinyl.lib.env import load_dotenv  # noqa: F401

__all__ = [
    "Field",
    "MetricStore",
    "VinylTable",
    "difference",
    "intersect",
    "join",
    "union",
    "Defs",
    "M",
    "T",
    "rich_print",
    "model",
    "resource",
    "source",
    "FillOptions",
    "types",
    "join",
]

install(
    suppress=[ibis, sqlglot, pydantic],
)


class M(MetricStore):  # noqa: F821
    pass


class T(VinylTable):  # noqa: F821
    pass


original_print = builtins.print  # type: ignore
builtins.print = rich_print  # type: ignore



import platform
import subprocess
import sys

from vinyl.lib.utils.files import _preserve_file_state


@_preserve_file_state("pyproject.toml")
def run_build():
    try:
        # Get the current platform
        current_platform = platform.system()

        if current_platform == "Windows":
            plat = "windows"

        elif current_platform == "Linux":
            plat = "linux"

        else:
            plat = "macos"

        current_version = sys.version_info
        version = f"cp{current_version.major}{current_version.minor}-*"

        subprocess.run(
            "poetry run python scripts/cache_project_scaffolding.py", shell=True
        )

        subprocess.run(
            "poetry run python scripts/generate_autocomplete_hints.py -p vinyl",
            shell=True,
        )
        subprocess.run(
            "poetry run python scripts/set_pyproject.py --engine nuitka", shell=True
        )
        subprocess.run(
            f"CIBW_BUILD={version} poetry run cibuildwheel --platform {plat} --output-dir dist/wheelhouse",
            shell=True,
        )
    finally:
        # make sure type stubs are removed no matter what
        subprocess.run(
            "poetry run python scripts/cleanup_autocomplete_hints.py", shell=True
        )


if __name__ == "__main__":
    run_build()



# type: ignore
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401

from internal_project.resources import local_duckdb  # noqa F401


@source(resource=local_duckdb)
class TaxiSample:
    _table = "taxi_sample"
    _unique_name = "local_duckdb.test.main.TaxiSample"
    _schema = "main"
    _database = "test"
    _twin_path = "data/test/main/local_duckdb.duckdb"
    _path = "../tests/vinyl/fixtures/test.duckdb"
    _row_count = 100000
    _col_replace = {}

    vendor_id: t.String(nullable=True)
    pickup_at: t.Timestamp(timezone=None, scale=6, nullable=True)
    dropoff_at: t.Timestamp(timezone=None, scale=6, nullable=True)
    passenger_count: t.Int8(nullable=True)
    trip_distance: t.Float32(nullable=True)
    rate_code_id: t.String(nullable=True)
    store_and_fwd_flag: t.String(nullable=True)
    pickup_location_id: t.Int32(nullable=True)
    dropoff_location_id: t.Int32(nullable=True)
    payment_type: t.String(nullable=True)
    fare_amount: t.Float32(nullable=True)
    extra: t.Float32(nullable=True)
    mta_tax: t.Float32(nullable=True)
    tip_amount: t.Float32(nullable=True)
    tolls_amount: t.Float32(nullable=True)
    improvement_surcharge: t.Float32(nullable=True)
    total_amount: t.Float32(nullable=True)
    congestion_surcharge: t.Float32(nullable=True)



import os
from enum import Enum

from vinyl import load_dotenv
from posthog import Posthog

from vinyl.cli.user import User

load_dotenv()


# Define your Event enum
class Event(Enum):
    LOGIN = "vinyl.login"
    SOURCE_GEN = " vinyl.source_gen"
    PROJECT_INIT = " vinyl.project_init"
    PREVIEW_MODEL = " vinyl.model.preview"
    PREVIEW_METRIC = " vinyl.metric.preview"
    DEPLOY = " vinyl.deploy"
    SERVE = " vinyl.serve"


class EventLogger:
    def __init__(self):
        if os.getenv("NO_POSTHOG_LOGGING") == "true":
            self.off = True
            return
        else:
            self.off = False
        self.user = User()
        self.posthog = Posthog(
            "phc_XL4KyheAjc4gJV4Fzpg1lbn7goFP1QNqsnNUhY1O1CU",
            host="https://us.posthog.com",
        )
        self.posthog.identify(
            self.user.user_info["user_id"],
            {"email": self.user.user_info["email"]},
        )

    def log_event(self, event: Event):
        if self.off:
            return
        self.posthog.capture(self.user.user_info["user_id"], event.value)



# type: ignore
from internal_project.resources import local_filesystem  # noqa F401
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401


@source(resource=local_filesystem)
class Stores:
    _table = "stores"
    _unique_name = "local_filesystem.Stores"
    _path = "data/csvs/stores.csv"
    _row_count = 54
    _col_replace = {}

    store_nbr: t.Int64(nullable=True) = Field(primary_key=True)
    city: t.String(nullable=True)
    state: t.String(nullable=True)
    type: t.String(nullable=True)
    cluster: t.Int64(nullable=True)



import functools
import importlib
import inspect
from functools import wraps
from typing import Any, Callable

from vinyl.lib.enums import AssetType
from vinyl.lib.metric import MetricStore
from vinyl.lib.settings import _get_project_module_name
from vinyl.lib.table import VinylTable
from vinyl.lib.utils.functions import _validate


@_validate
def _base(
    deps: object | Callable[..., Any] | list[object | Callable[..., Any]],
    asset_type: AssetType | None = None,
    read_only: bool = False,
    publish: bool = False,
    tags: str | list[str] | None = None,
    description: str | None = None,
    column_descriptions: dict[str, str] | None = None,
):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args):
            if not isinstance(deps, list):
                deps_adj = [deps]
            else:
                deps_adj = deps
            # Map the positional arguments to the new names
            params = [
                param.name for param in inspect.signature(func).parameters.values()
            ]
            if len(params) != len(deps_adj):
                raise Exception("Wrong number of arguments")

            new_kwargs = {}
            # do dependency injection for each param of the function

            for i, param in enumerate(params):
                if deps_adj[i] is None:
                    continue
                # TODO: this is a hack to avoid mutable sources and models from being impacted downstream
                # we should eventually clean this up as it's not a sustainable solution
                dep_it = deps_adj[i]()
                if isinstance(dep_it, (VinylTable, MetricStore)):
                    par = dep_it._copy()

                else:
                    raise ValueError(
                        f"Dependencies must be VinylTable or MetricStore, not {type(dep_it)}"
                    )

                par._mutable = True
                new_kwargs[param] = par

            return func(**new_kwargs)

        setattr(wrapper, "_module", inspect.getmodule(func).__name__)
        setattr(wrapper, f"_is_vinyl_{asset_type.value}", True)
        if asset_type == AssetType.MODEL or AssetType.METRIC:
            deps_helper = [deps] if not isinstance(deps, list) else deps
            setattr(wrapper, "_deps", deps_helper)
        if tags:
            setattr(wrapper, "_tags", tags)
        if description:
            setattr(wrapper, "_description", description)
        if publish:
            setattr(wrapper, "_publish", publish)

        return wrapper

    return decorator


@_validate
def model(
    deps: object | Callable[..., Any] | list[object | Callable[..., Any]],
    read_only: bool = False,
    publish: bool = False,
    destination: list[str] | None = None,
    tags: str | list[str] | None = None,
    description: str | None = None,
    column_descriptions: dict[str, str] | None = None,
):
    return _base(
        deps=deps,
        asset_type=AssetType.MODEL,
        read_only=read_only,
        publish=publish,
        tags=tags,
        description=description,
        column_descriptions=column_descriptions,
    )


@_validate
def metric(
    deps: object | Callable[..., Any] | list[object | Callable[..., Any]],
    tags: str | list[str] | None = None,
    publish: bool = False,
    description: str | None = None,
    metric_descriptions: dict[str, str] | None = None,
):
    return _base(
        deps=deps,
        asset_type=AssetType.METRIC,
        publish=publish,
        tags=tags,
        description=description,
        column_descriptions=metric_descriptions,
    )


def resource(func: Callable[..., Any], default: bool = False):
    @wraps(func)
    def decorated_function(*args, **kwargs):
        return func(*args, **kwargs)

    decorated_function._is_vinyl_resource = True  # type: ignore
    decorated_function._default = default  # type: ignore

    return decorated_function


def get_resource_by_name(name: str):
    module = importlib.import_module(f"{_get_project_module_name()}.resources")
    return getattr(module, name)



from ibis import Schema
from vinyl.lib.utils.graphics import _print_schema


class VinylSchema:
    _schema: Schema

    def __init__(self, schema: Schema):
        self._schema = schema

    def __str__(self):
        return self._schema.__str__().replace("ibis.Schema {", "vinyl.Schema {")

    def __repr__(self):
        return self._schema.__repr__().replace("ibis.Schema {", "vinyl.Schema {")

    def items(self):
        return self._schema.items()

    def get(self, name: str):
        return self._schema.get(name)

    @property
    def names(self) -> list[str]:
        return self._schema.names

    @property
    def types(self) -> list[str]:
        return self._schema.types

    def __rich__(self):
        return _print_schema(self._schema)

    def to_dict(self):
        return {name: str(type) for name, type in self._schema.items()}



import copy
import email.utils

import toml


def set_pyproject():
    with open("pyproject.toml", "r") as f:
        pyproject = toml.load(f)

    poetry = copy.deepcopy(pyproject["tool"]["poetry"])
    build_system = pyproject["build-system"]
    build_system["requires"] = ["setuptools>=42", "wheel", "nuitka", "toml"]
    build_system["build-backend"] = "nuitka.distutils.Build"

    if "project" not in pyproject:
        pyproject["project"] = {}
    proj = pyproject["project"]

    proj["name"] = poetry["name"]
    proj["version"] = poetry["version"]
    proj["description"] = poetry["description"]
    adj_authors = [email.utils.parseaddr(auth) for auth in poetry["authors"]]
    proj["authors"] = [{"name": auth[0], "email": auth[1]} for auth in adj_authors]
    proj["scripts"] = poetry["scripts"]
    proj["license"] = {"file": poetry["license"]}
    proj["requires-python"] = poetry["dependencies"]["python"]
    proj["readme"] = poetry["readme"]
    proj["classifiers"] = poetry["classifiers"]

    proj["dynamic"] = ["dependencies"]

    with open("pyproject.toml", "w") as f:
        toml.dump(pyproject, f)


if __name__ == "__main__":
    set_pyproject()



import sqlglot as sg
import sqlglot.expressions as sge
import sqlglot.optimizer as sgo
from sqlglot.optimizer.qualify import qualify


def get_column_completion(
    sql_statement: str, table_or_cte: str, dialect: str = "duckdb"
):
    parsed = sg.parse(sql_statement, dialect=dialect)[0]
    model_ast = parsed
    if model_ast is None:
        return []

    sources = []
    for x in model_ast.find_all(sge.CTE):
        # split sources if there are unions
        sources += [(x.alias, x.this)]
        x.pop()
    for x2 in model_ast.find_all(sge.With):
        x2.pop()
    sources += [("test", model_ast)]
    adj = sg.exp.Select(
        **{
            "expressions": [sge.Star()],
            "from": sge.From(
                this=sge.Table(this=sge.Identifier(this="test", quoted=False))
            ),
            "with": sge.With(
                expressions=[
                    sge.CTE(
                        alias=sge.TableAlias(
                            this=sge.Identifier(this=source[0], quoted=False)
                        ),
                        this=source[1],
                    )
                    for source in sources
                ]
            ),
        }
    )
    optimized = sgo.optimize(expression=adj, rules=[qualify])
    return [c.alias for c in optimized.expressions]


if __name__ == "__main__":
    sql = "with x as (SELECT t.three FROM my_table as t), y as (select a.two from pear as a) select y.* from y left join x"
    # sql = "with x as (SELECT t.three FROM my_table as t) select * from x"
    out = get_column_completion(sql, "t")



from contextlib import asynccontextmanager
from typing import List

from fastapi import Depends, FastAPI, Query, Request
from vinyl.lib.project import Project
from vinyl.lib.query_engine import QueryEngine


class ProjectDependency:
    def __init__(self, project: Project):
        self.project = project


@asynccontextmanager
async def lifespan(app: FastAPI):
    # load project on startup
    app.state.project = Project.bootstrap()
    yield
    # clean up here
    pass


def get_project(request: Request) -> Project:
    return request.app.state.project


# Create an instance of the FastAPI class
app = FastAPI(lifespan=lifespan)


# Define a route at the root '/'
@app.get("/")
async def _root():
    return {"status": "OK"}


@app.get("/models/{name}")
async def _get_model(
    name: str,
    limit: int = 1000,
    project: ProjectDependency = Depends(get_project),
):
    query_engine = QueryEngine(project)
    result = query_engine._model(name, limit)
    return {"limit": limit, "rows": result.to_records()}


@app.get("/metrics/{name}")
async def _get_metrics(
    name: str,
    grain: str,
    metrics: List[str] = Query([]),
    dimensions: List[str] = Query([]),
    limit: int = 1000,
    project: ProjectDependency = Depends(get_project),
):
    """
    Fetch metrics based on the metric name, time granularity, and dimensions

    Args:
    - metric_name (str): The name of the metric.
    - time_grain (str): The granularity of time (e.g., hourly, ex: days=14).
    - dimensions (List[str]): An array of dimensions to filter the metrics.

    Returns:
    - dict: A pandas records with the query parameters received.
    """
    query_engine = QueryEngine(project)
    result = query_engine._metric(store=name, grain=grain, limit=limit)
    return {
        "limit": limit,
        "rows": result.to_records(),
    }



from __future__ import annotations

import dataclasses
import traceback
from typing import Any, Callable

import ibis.expr.types as ir
from mpire import WorkerPool
from rich.table import Table
from sqlglot.errors import ParseError
from tqdm import tqdm

from vinyl.lib.connect import SourceInfo, _ResourceConnector
from vinyl.lib.definitions import Defs, _load_project_defs
from vinyl.lib.errors import VinylError, VinylErrorType
from vinyl.lib.metric import MetricStore
from vinyl.lib.sqlast import SQLAstNode, SQLProject, get_adj_node_name
from vinyl.lib.table import VinylTable
from vinyl.lib.utils.graph import DAG
from vinyl.lib.utils.graphics import rich_print


@dataclasses.dataclass
class Resource:
    name: str
    connector: _ResourceConnector
    def_: Any


class Project:
    resources: list[Resource]
    sources: list[ir.Table]
    models: list[Callable[..., Any]]
    metrics: list[Callable[..., Any]]
    graph: DAG
    db_graph: DAG
    default_resource: Resource

    def __init__(
        self,
        resources: list[Callable[..., Any]],
        sources: list[ir.Table] | None = None,
        models: list[Callable[..., Any]] | None = None,
        metrics: list[Callable[..., Any]] | None = None,
        graph: DAG | None = None,
        db_graph: DAG | None = None,
        default_resource: Resource | None = None,
    ):
        self.resources = [
            Resource(
                name=getattr(resource_def, "_name_override", resource_def.__name__),
                def_=resource_def,
                connector=resource_def(),
            )
            for resource_def in resources
        ]
        if sources is not None:
            self.sources = sources
        if models is not None:
            self.models = models
        if metrics is not None:
            self.metrics = metrics

        self.graph = DAG() if graph is None else graph
        self.db_graph = DAG() if db_graph is None else db_graph
        self.default_resource = (
            self.resources[0] if default_resource is None else default_resource
        )

    def _get_source_objects(self, with_schema=False) -> list[SourceInfo]:
        sources = []
        errors = []
        for resource in self.resources:
            try:
                resource_sources, errors_it = resource.connector._list_sources(
                    with_schema
                )
                for source in resource_sources:
                    source._parent_resource = resource
                    sources.append(source)
                errors.extend(errors_it)
            except Exception as e:
                print(f"Error loading sources from {resource.name}: {e}")
                errors.append(
                    VinylError(
                        resource.name,
                        type=VinylErrorType.DATABASE_ERROR,
                        msg=str(e),
                        traceback=traceback.format_exc(),
                    )
                )
        return sources, errors

    def _get_resource(self, resource_name: str) -> Resource:
        resources = [
            resource for resource in self.resources if resource.name == resource_name
        ]

        if len(resources) == 0:
            raise ValueError(f"Resource {resource_name} not found")

        resource = resources[0]

        return resource

    def _get_source(self, source_id: str):
        sources = [source for source in self.sources if source.__name__ == source_id]

        if len(sources) == 0:
            raise ValueError(f"Source {source_id} not found")

        source = sources[0]

        return source

    def _get_model(self, model_id: str) -> VinylTable:
        if self.models is None:
            raise ValueError("No models found")
        models = [model for model in self.models if model.__name__ == model_id]

        if len(models) == 0:
            raise ValueError(f"Model {model_id} not found")

        model = models[0]

        return model()

    def _get_metric_store(self, metric_id: str) -> MetricStore:
        metrics = [metric for metric in self.metrics if metric.__name__ == metric_id]
        if len(metrics) == 0:
            raise ValueError(f"Metric {metric_id} not found")

        metric = metrics[0]

        return metric()

    def _select_lineage_nodes(
        self,
        ids: list[str] | None = None,
        predecessor_depth: int | None = None,
        successor_depth: int | None = None,
    ):
        if ids is None:
            nodes = self.db_graph.topological_sort()
        else:
            node_id_map = {
                get_adj_node_name(n): n for n in self.graph.node_dict if n is not None
            }
            nodes_to_consider = [node_id_map[id] for id in ids]
            all_nodes_to_consider = self.db_graph.get_ancestors_and_descendants(
                nodes_to_consider, predecessor_depth, successor_depth
            )
            nodes = self.db_graph.subgraph(
                list(all_nodes_to_consider)
            ).topological_sort()

        nodes = [
            n for n in nodes if not isinstance(n, VinylTable)
        ]  # remove source nodes from lineage calc

        return nodes

    def get_sql_project(
        self,
        ids: list[str] | None = None,
        predecessor_depth: int | None = None,
        successor_depth: int | None = None,
        calculate_lineage: bool = True,
        parallel: bool = False,
    ):
        nodes = self._select_lineage_nodes(ids, predecessor_depth, successor_depth)

        def process_node(shared_objects, node):
            graph = shared_objects[0]
            sqlnode = SQLAstNode()
            try:
                sqlnode.id = get_adj_node_name(node)
                relative_to = {}
                for y in graph.get_parents([node]):
                    if y is None:
                        continue

                    try:
                        y_name = get_adj_node_name(y)
                        active_y = y()
                        relative_to[y_name] = active_y
                    except Exception as e:
                        sqlnode.errors.append(
                            VinylError(
                                node_id=sqlnode.id,
                                source_id=get_adj_node_name(y),
                                msg=str(e),
                                type=VinylErrorType.MISCELLANEOUS_ERROR,
                                traceback=traceback.format_exc(),
                            )
                        )
                        # don't fail here, fail later if there are no deps to be found. Swallowing the exception allows for partial lineage
                        continue
                sqlnode.ast, sqlnode.schema, sqlnode.dialect = node()._to_sql_ast(
                    relative_to=relative_to
                )
                sqlnode.deps = [k for k in relative_to]
                sqlnode.deps_schemas = {k: v.schema() for k, v in relative_to.items()}

            except Exception as e:
                sqlnode.errors.append(
                    VinylError(
                        node_id=sqlnode.id,
                        type=(
                            VinylErrorType.PARSE_ERROR
                            if isinstance(e, ParseError)
                            else VinylErrorType.MISCELLANEOUS_ERROR
                        ),
                        dialect=sqlnode.dialect,
                        msg=str(e),
                        traceback=traceback.format_exc(),
                        context=str(node),
                    )
                )
            return sqlnode

        if parallel:
            with WorkerPool(
                n_jobs=min(5, len(nodes)),
                use_dill=True,
                start_method="spawn",
                shared_objects=[self.db_graph],
                keep_alive=True,
            ) as pool:
                sqlnodelist = pool.map(process_node, nodes, progress_bar=True)
        else:
            sqlnodelist = [process_node([self.db_graph], node) for node in tqdm(nodes)]

        adj_db_graph = self.db_graph.copy().relabel(
            lambda x: get_adj_node_name(x) if x is not None else x
        )

        return SQLProject(sqlnodelist, adj_db_graph)

    def list_sources(self, tables: bool = False):
        table = Table("Name", "Resource", "Location", title="Sources")
        for source in self._get_source_objects()[0]:
            if source._parent_resource is None:
                raise ValueError("Source must have a parent resource.")
            table.add_row(
                f"[bold]{source._name}[bold]",
                f"[grey70]{source._parent_resource.def_.__name__}[grey70]",
                f"[grey70]{source._location}[grey70]",
            )
        rich_print(table)

    @classmethod
    def load(cls, defs: Defs) -> Project:
        return cls(
            resources=defs.resources,
            sources=defs.sources,
            models=defs.models,
            metrics=defs.metrics,
            graph=defs.graph,
            db_graph=defs.db_graph,
            default_resource=defs.default_resource,
        )

    @classmethod
    def bootstrap(cls) -> Project:
        defs = _load_project_defs()
        return cls.load(defs)



from datetime import datetime
from typing import Any, Callable

import ibis
import ibis.expr.types as ir
import pytz
from ibis.common.deferred import Deferred

from vinyl.lib.column_methods import ColumnBuilder
from vinyl.lib.settings import PyProjectSettings


def _get_utc_offset(timezone_str: str) -> float:
    # Create a timezone object
    tz = pytz.timezone(timezone_str)

    # Localize a datetime object to the specified timezone
    localized_dt = tz.localize(datetime.now())

    # Get the UTC offset in minutes
    utc_offset = localized_dt.utcoffset()
    if utc_offset is not None:
        return utc_offset.total_seconds() / 60

    raise ValueError(f"Invalid timezone: {timezone_str}")


## NOTE: This is a bit hacky. Some databases don't support timestamps with timezones effectively, so we use adjusted timestamps without timezones instead.
def _set_timezone(
    tbl: Any,  # will be vinylTable but can't import here due to circular reference
    col: ir.Value | Callable[..., Any],
    to_tz: str | None = None,
    from_tz: str = "UTC",
) -> Deferred | ir.Value:
    # Check if col is a timestamp, otherwise raise an error
    vinyl_col = ColumnBuilder(tbl, col)

    if not vinyl_col._type or not vinyl_col._type.is_timestamp():
        raise TypeError("Column must be of type timestamp")

    # If no timezone is specified, check pyproject.toml
    to_tz = (
        PyProjectSettings()._get_setting("tz")
        or str(datetime.now().astimezone().tzinfo)
        or "UTC"
    )

    # Create a timezone object
    utc_offset = _get_utc_offset(to_tz)
    if from_tz != "UTC":
        utc_offset -= _get_utc_offset(from_tz)

    # Cast to timestamp without tz (a bit hacky in ibis) and add the new timezone. Note this is a bit hacky in ibis. You can't cast to an instance of dt.Timestamp of it is deferred.
    col_as_dt = (
        col.cast(datetime) + ibis.interval(minutes=utc_offset)
        if isinstance(col, ir.Value) or isinstance(col, Deferred)
        else lambda x: col(x.tbl).cast(datetime) + ibis.interval(minutes=utc_offset)
    )
    return col_as_dt



from __future__ import annotations

from collections import Counter

import ibis
import ibis.expr.operations as ops
import ibis.expr.types as ir
import ibis.selectors as s
import networkx as nx
import rustworkx as rx
from ibis.backends import BaseBackend
from ibis.common.collections import FrozenDict
from rich.console import Console
from rich.text import Text  # noqa: F401
from textual.app import ComposeResult
from textual.widgets import Footer
from vinyl.lib.utils.context import _is_notebook
from vinyl.lib.utils.graph import _get_label, _rustworkx_to_networkx
from vinyl.lib.utils.graphics import TurntableTextualApp


def _find_backend(node: ops.Node) -> BaseBackend | None:
    nodes_sources = (ops.DatabaseTable, ops.SQLQueryResult, ops.UnboundTable)
    if isinstance(node, nodes_sources):
        return node.source
    pos_nodes = node.find(nodes_sources)
    if len(pos_nodes) > 0:
        return pos_nodes[0].source

    return None


def _leafs(g: rx.PyDiGraph) -> list[int]:
    return [index for index in g.node_indexes() if g.out_degree(index) == 0]


def _roots(g: rx.PyDiGraph) -> list[int]:
    return [index for index in g.node_indexes() if g.in_degree(index) == 0]


class VinylGraph(rx.PyDiGraph):
    def _leafs(self):
        return _leafs(self)

    def _roots(self):
        return _roots(self)

    @classmethod
    def _new_init_from_expr(cls, expr, label_func=None):
        return VinylGraph._init_from_expr_helper(
            expr.op(), VinylGraph(), {}, label_func
        )

    @staticmethod
    def _init_from_expr_helper(
        u, g: VinylGraph, node_dict: dict[str, int], label_func=None
    ):
        if isinstance(u, ops.Node):
            u_hash = str(u.__hash__())
            if u_hash not in node_dict:
                node_dict[u_hash] = g.add_node(
                    u if label_func is None else label_func(u)
                )
            u_index = node_dict[u_hash]
            for v in u.__children__:
                if isinstance(v, ops.Node):
                    v_hash = str(v.__hash__())
                    if v_hash not in node_dict:
                        node_dict[v_hash] = g.add_node(
                            v if label_func is None else label_func(v)
                        )
                    v_index = node_dict[v_hash]
                    if (u_index, v_index) not in g.edges():
                        g.add_edge(u_index, v_index, 1)
                    g = VinylGraph._init_from_expr_helper(v, g, node_dict, label_func)

        node_dict = {}  # reset to ensure no future issues
        return g

    @staticmethod
    def _find_unaltered_cols(expr) -> dict[ir.Expr, ir.Expr]:
        g = VinylGraph._new_init_from_expr(expr.select(s.all()).tbl)

        # delete edge_to_edge connection
        for u, v in g.edge_list():
            if isinstance(g[u], ops.Relation) and isinstance(g[v], ops.Relation):
                g.remove_edge(u, v)

        # delete final table on either side of the graph:
        for leaf in g._leafs():
            g.remove_node(leaf)

        for root in g._roots():
            g.remove_node(root)

        # store the end nodes of the final table
        columns_to_check = g._leafs()

        # remove the relation nodes and reconnect them to the final table if the column name is the same
        for node in g.node_indexes():
            if isinstance(g[node], ops.Relation):
                frozens = [arg for arg in g[node].args if isinstance(arg, FrozenDict)]
                combined = {}
                for f in frozens:
                    for k, v in f.items():
                        combined[k] = v

                in_nodes = list(g.predecessor_indices(node))
                out_nodes = list(g.successor_indices(node))
                for in_node in in_nodes:
                    for out_node in out_nodes:
                        if combined[g[in_node].name] == g[out_node]:
                            g.add_edge(in_node, out_node, 1)
                g.remove_node(node)

        unaltered = {}
        orig_g = g.copy()
        for col in columns_to_check:
            sub = orig_g.subgraph([col, *rx.ancestors(g, col)])  # type: ignore[attr-defined]
            root = _roots(sub)
            if len(root) > 1:
                continue
            elif all(
                [
                    isinstance(sub[i], (ops.Alias, ops.Field, ops.SortKey))
                    for i in sub.node_indexes()
                ]
            ):
                unbound_tbl, col_name = orig_g[col].args
                unaltered[getattr(unbound_tbl.to_expr(), col_name)] = sub[root[0]]

        return unaltered

    @classmethod
    def _visualize(cls, expr: ir.Expr):
        from netext import ConsoleGraph
        from netext.edge_rendering.arrow_tips import ArrowTip
        from netext.edge_rendering.modes import EdgeSegmentDrawingMode
        from netext.edge_routing.modes import EdgeRoutingMode
        from netext.textual.widget import GraphView

        g_rust = cls._new_init_from_expr(expr, label_func=lambda x: x)
        g = _rustworkx_to_networkx(
            g_rust,
        )

        nx.set_node_attributes(
            g,
            {n: _get_label(n, markup=True) for i, n in enumerate(g.nodes())},
            "$label",
        )

        nx.set_node_attributes(g, lambda n, d, s: d["$label"], "$content-renderer")
        nx.set_edge_attributes(g, EdgeRoutingMode.ORTHOGONAL, "$edge-routing-mode")
        nx.set_edge_attributes(
            g, EdgeSegmentDrawingMode.BOX_ROUNDED, "$edge-segment-drawing-mode"
        )
        nx.set_edge_attributes(g, ArrowTip.ARROW, "$start-arrow-tip")

        if _is_notebook():
            console = Console(
                width=10000
            )  # make width large to avoid jupyter cutting off output
            console.print(ConsoleGraph(g, zoom=1.2, console=console))

        else:

            class GraphviewApp(TurntableTextualApp):
                def compose(self) -> ComposeResult:
                    yield GraphView(g, zoom=1.2)
                    yield Footer()

            GraphviewApp().run()


def _unify_backends(tbl, visited_nodes=set()):
    g = VinylGraph._new_init_from_expr(tbl)
    for i in rx.topological_sort(g):
        node = g[i]
        if node in visited_nodes:
            continue
        visited_nodes.add(node)

        if isinstance(
            node, (ops.JoinChain, ops.Union, ops.Difference, ops.Intersection)
        ):
            if isinstance(node, ops.JoinChain):
                nodes = [node.first.parent]
                nodes.extend([n.table.parent for n in node.rest])
            else:
                nodes = [node.left, node.right]
            sources = [_find_backend(n) for n in nodes]
            non_null_sources = [s for s in sources if s is not None]

            # determine dominant backend
            counted = Counter(sources)
            most_common = counted.most_common()

            # only do replacement if left and right have different sources and neither is a MemTable
            if len(non_null_sources) > 0 and len(set(non_null_sources)) > 1:
                expressions = [n.to_expr() for n in nodes]
                schemas = [n.schema for n in nodes]
                replace = {}

                # if one of the most common sources is duckdb, use that
                if any(
                    [
                        isinstance(n[0], ibis.backends.duckdb.Backend)
                        for n in most_common
                    ]
                ):
                    for i, node in enumerate(nodes):
                        if not isinstance(sources[i], ibis.backends.duckdb.Backend):
                            replace[node] = ibis.memtable(
                                data=expressions[i].to_pyarrow(), schema=schemas[i]
                            ).op()

                # otherwise use the first most common source:
                else:
                    for i, node in enumerate(nodes):
                        if sources[i] != most_common[0][0]:
                            replace[node] = ibis.memtable(
                                data=expressions[i].to_pyarrow(), schema=schemas[i]
                            ).op()

                tbl = tbl.op().replace(replace).to_expr()
                return _unify_backends(tbl, visited_nodes)

    return tbl



# used for autocomplete for VinylTable and MetricStore classes
from __future__ import annotations

from collections.abc import Iterable, Sequence
from functools import wraps
from typing import Any, Callable, Literal

import ibis
import ibis.expr.operations as ops
from ibis.common.deferred import Deferred
from ibis.expr.datatypes import DataType
from ibis.expr.types import (
    ArrayColumn,
    ArrayValue,
    BooleanValue,
    Column,
    DateValue,
    Expr,
    IntegerValue,
    IntervalValue,
    MapColumn,
    NumericValue,
    Scalar,
    StringColumn,
    StringValue,
    StructColumn,
    TimestampValue,
    Value,
)


# self excluded to avoid recursion
def _promote_output(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        if isinstance(result, Value):
            return VinylColumn(result)
        return result

    return wrapper


def _promote_output_outside_class(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if isinstance(result, Value):
            return VinylColumn(result)
        return result

    return wrapper


def _adj_args_and_output(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        new_args = []
        for i, arg in enumerate(args):
            new_args.append(_demote_arg(arg))
        new_kwargs = {}
        for k, v in kwargs.items():
            new_kwargs[k] = _demote_arg(v)
        return func(self, *new_args, **new_kwargs)

    return wrapper


# Class decorator that applies the function decorator to all methods
def _transform_all_methods(cls):
    for attr_name, attr_value in cls.__dict__.items():
        if callable(attr_value):
            setattr(cls, attr_name, _promote_output(_adj_args_and_output(attr_value)))
    return cls


@_transform_all_methods
class VinylColumn:
    _col: Value
    array: VinylColumn.ArrayFunctions
    math: VinylColumn.MathFunctions
    url: VinylColumn.URLFunctions
    re: VinylColumn.RegexFunctions
    str: VinylColumn.StringFunctions
    dict: VinylColumn.MapFunctions
    obj: VinylColumn.StructFunctions
    dt: VinylColumn.TemporalFunctions

    ## dunders not implemented: __complex__, __format__, __getslice__ (deprecated), __index__, __matmul__, __rmatmul__
    ## dunders that could be implemented but not yet: __reversed__
    def __init__(self, _col: Column):
        self._col = _col
        self.array = self.ArrayFunctions()  # type: ignore[call-arg]
        self.math = self.MathFunctions()  # type: ignore[call-arg]
        self.url = self.URLFunctions()  # type: ignore[call-arg]
        self.re = self.RegexFunctions()  # type: ignore[call-arg]
        self.str = self.StringFunctions()  # type: ignore[call-arg]
        self.dict = self.MapFunctions()  # type: ignore[call-arg]
        self.obj = self.StructFunctions()  # type: ignore[call-arg]
        self.dt = self.TemporalFunctions()  # type: ignore[call-arg]

    def __repr__(self) -> str:  # type: ignore[valid-type]
        return self._col.__repr__()

    def __str__(self) -> str:  # type: ignore[valid-type]
        return self._col.__str__()

    # not implemented by ibis, but adding here
    def __abs__(self) -> VinylColumn:
        return self._col.abs()

    def __add__(self, other: Any) -> VinylColumn:
        return type(self._col).__add__(self._col, other)

    def __and__(self, other: Any) -> VinylColumn:
        if isinstance(self._col, ArrayColumn):
            return self._col.intersect(other)
        return type(self._col).__and__(self._col, other)

    def __bool__(self) -> VinylColumn:
        return type(self._col).__bool__(self._col)

    # not implemented by ibis, but adding here
    def __ceil__(self) -> VinylColumn:
        return self._col.ceil()

    def __cmp__(self, other: Any) -> VinylColumn:
        # using ibis ifelse here to avoid recursion
        return ibis.ifelse(self._col == other, 0, ibis.ifelse(self._col < other, -1, 1))

    def __contains__(self, key: Any) -> VinylColumn:
        return type(self._col).__contains__(self._col, key)

    def __divmod__(self, other: Any) -> VinylColumn:
        return type(self._col).__divmod__(self._col, other)

    def __eq__(self, other: Any) -> VinylColumn:  # type: ignore
        if other is None:
            return self._col.isnull()
        return type(self._col).__eq__(self._col, other)  # type: ignore

    # not implemented by ibis, but adding here
    def __float__(self) -> VinylColumn:
        return self._col.cast(float)

    # not implemented by ibis, but adding here
    def __floor__(self) -> VinylColumn:
        return self._col.floor()

    def __floordiv__(self, other: Any) -> VinylColumn:
        return type(self._col).__floordiv__(self._col, other)

    def __ge__(self, other: Any) -> VinylColumn:
        return type(self._col).__ge__(self._col, other)

    def __getitem__(self, key: Any) -> VinylColumn:
        if not isinstance(self._col, StringColumn):
            return type(self._col).__getitem__(self._col, key)

        if isinstance(key, slice):
            start, stop, step = key.start, key.stop, key.step

            if not isinstance(start, (Expr, VinylColumn)):
                if start is not None and start < 0:
                    start = self._col.length() + start

            if not isinstance(stop, (Expr, VinylColumn)):
                if stop is not None and stop < 0:
                    stop = self._col.length() + stop

            if step == -1:
                adj_stop = self._col.length() - start - 1 if start is not None else None
                adj_start = self._col.length() - stop - 1 if stop is not None else None
                return VinylColumn(self._col.reverse()).__getitem__(
                    slice(adj_start, adj_stop, 1)
                )

            if start is None:
                start = 0
            if stop is None:
                stop = self._col.length()

                if stop is None:
                    stop = self._col.length()

            if step is not None and not isinstance(step, Expr) and abs(step) != 1:
                raise ValueError("Step can only be 1 or -1 for string slicing")

            return self.str.substr(start, stop - start)

        elif isinstance(key, int):
            return self.str.substr(key, 1)
        raise NotImplementedError(f"string __getitem__[{key.__class__.__name__}]")

    def __gt__(self, other: Any) -> VinylColumn:
        return type(self._col).__gt__(self._col, other)

    def __hash__(self) -> VinylColumn:  # type: ignore
        return type(self._col).__hash__(self._col)  # type: ignore

    # not implemented by ibis, but adding here
    def __int__(self) -> VinylColumn:
        return self._col.cast(int)

    def __invert__(self) -> VinylColumn:
        return type(self._col).__invert__(self._col)

    def __iter__(self) -> VinylColumn:
        return type(self._col).__iter__(self._col)

    def __le__(self, other: Any) -> VinylColumn:
        return type(self._col).__le__(self._col, other)

    # doesn't really work, because python requires len to return an integer
    def __len__(self) -> VinylColumn:
        return type(self._col).__len__(self._col)

    def __lshift__(self, other: Any) -> VinylColumn:
        return type(self._col).__lshift__(self._col, other)

    def __lt__(self, other: Any) -> VinylColumn:
        return type(self._col).__lt__(self._col, other)

    def __mod__(self, other: Any) -> VinylColumn:
        return type(self._col).__mod__(self._col, other)

    def __mul__(self, other: Any) -> VinylColumn:
        return type(self._col).__mul__(self._col, other)

    def __ne__(self, other: Any) -> VinylColumn:  # type: ignore
        if other is None:
            return self._col.notnull()
        return type(self._col).__ne__(self._col, other)  # type: ignore

    def __neg__(self) -> VinylColumn:
        # helps with sorting behavior
        if not hasattr(type(self._col), "__neg__"):
            # turn this into a negated sort key
            return ops.SortKey(self._col.op(), ascending=False).to_expr()

        # sort key conversion will be handled later, and this prevents issues with negating numerics, etc.
        return type(self._col).__neg__(self._col)

    # likely not implemented, but including just in case
    def __next__(self) -> VinylColumn:
        return type(self._col).__next__(self._col)

    def __nonzero__(self) -> VinylColumn:
        return type(self._col).__nonzero__(self._col)

    def __or__(self, other: Any) -> VinylColumn:  # type: ignore
        if isinstance(self._col, ArrayColumn) and isinstance(other, ArrayColumn):
            return self._col.union(other)
        return type(self._col).__or__(self._col, other)  # type: ignore

    # not implemented by ibis, but adding here
    def __pos__(self) -> VinylColumn:
        return self._col

    def __pow__(self, other: Any) -> VinylColumn:
        return type(self._col).__pow__(self._col, other)

    def __radd__(self, other: Any) -> VinylColumn:
        return type(self._col).__radd__(self._col, other)

    def __rand__(self, other: Any) -> VinylColumn:
        return type(self._col).__rand__(self._col, other)

    def __rdiv__(self, other: Any) -> VinylColumn:
        return type(self._col).__rdiv__(self._col, other)

    def __rdivmod__(self, other: Any) -> VinylColumn:
        return type(self._col).__rdivmod__(self._col, other)

    def __rfloordiv__(self, other: Any) -> VinylColumn:
        return type(self._col).__rfloordiv__(self._col, other)

    def __rlshift__(self, other: Any) -> VinylColumn:
        return type(self._col).__rlshift__(self._col, other)

    def __rmul__(self, other: Any) -> VinylColumn:
        return type(self._col).__rmul__(self._col, other)

    def __ror__(self, other: Any) -> VinylColumn:  # type: ignore
        return type(self._col).__ror__(self._col, other)  # type: ignore

    # not implemented by ibis, but adding here
    def __round__(self, ndigits: Any) -> VinylColumn:
        return self._col.round(ndigits)

    def __rpow__(self, other: Any) -> VinylColumn:
        return type(self._col).__rpow__(self._col, other)

    def __rrshift__(self, other: Any) -> VinylColumn:
        return type(self._col).__rrshift__(self._col, other)

    def __rshift__(self, other: Any) -> VinylColumn:
        return type(self._col).__rshift__(self._col, other)

    def __rsub__(self, other: Any) -> VinylColumn:
        return type(self._col).__rsub__(self._col, other)

    def __rtruediv__(self, other: Any) -> VinylColumn:
        return type(self._col).__rtruediv__(self._col, other)

    def __rxor__(self, other: Any) -> VinylColumn:
        return type(self._col).__rxor__(self._col, other)

    def __sub__(self, other: Any) -> VinylColumn:
        return type(self._col).__sub__(self._col, other)

    def __truediv__(self, other: Any) -> VinylColumn:
        return type(self._col).__truediv__(self._col, other)

    # not implemented by ibis, but adding here
    def __trunc__(self):
        return ibis.ifelse(self._col < 0, self._col.floor(), self._col.ceil())

    def __xor__(self, other: Any) -> VinylColumn:
        return type(self._col).__xor__(self._col, other)

    def between(self, lower: Value, upper: Value) -> VinylColumn:
        """
        Check if this expression is between lower and upper, inclusive.
        """
        return VinylColumn(self._col.between(lower, upper))

    def match(
        self, case_pairs: list[tuple[BooleanValue, Value]], default: Value | None = None
    ) -> Value:
        """
        Return a value based on the first matching condition in the expression. The default value is returned if no conditions are met, otherwise null is returned.
        """
        out = self._col.case()
        for pair in case_pairs:
            out = out.when(pair[0], pair[1])
        if default is not None:
            out = out.else_(default)

        return out

    def cast(self, target_type: Any, try_: bool = False) -> Value:
        """
        Cast expression to indicated data type. Type inputs can include strings, python type annotations, numpy dtypes, pandas dtypes, and pyarrow dtypes.

        If try_ is True, then the cast will return null if the cast fails, otherwise it will raise an error.
        """
        if try_:
            return self._col.try_cast(target_type)
        return self._col.cast(target_type)

    def coalesce(self, *args: Value) -> Value:
        """
        Return the first non-null value in the expression list.
        """
        return ibis.coalesce(self._col, *args)

    def fillna(self, value: Value) -> Value:
        """
        Replace null values in the expression with the specified value.
        """
        return self._col.fillna(value)

    def hash(self) -> IntegerValue:
        """
        Compute an integer hash value of the expression.

        The hashing function used is dependent on the backend, so usage across dialect will likely return a different number.
        """
        return self._col.hash()

    def equivalent(self, other: Value) -> BooleanValue:
        """
        Null-aware version of ==. Returns true if both expressions are equal or both are null.
        """
        return self._col.identical_to(other)

    def isin(self, values: Value | Sequence[Value]) -> BooleanValue:
        """
        Check if this expression is in the provided set of values. Exists in place of the python `in` operator because of its requirement to evaluate to a python boolean.
        """
        return self._col.isin(values)

    def type(self, db_type: bool = True) -> DataType | IntegerValue:
        """
        Return the string name of the datatype of the expression.

        If db_type is True, then the string will be the name of the datatype in the specific backend (e.g. duckdb), otherwise it will be cross-dialect data type name from Vinyl.
        """
        if db_type:
            return self._col.typeof()
        return self._col.type()

    def median(self, where: BooleanValue | None = None, approx: bool = False) -> Scalar:
        """
        Return the median value of the expression.

        If a `where` condition is specified, method only considers rows meeting the `where` condition.

        If `approx` is True, method will use the approximate median function, which is faster but less accurate.
        """
        if approx:
            return self._col.approx_median(where)
        return self._col.median(where)

    def count(
        self,
        where: BooleanValue | None = None,
        distinct: bool = False,
        approx: bool = False,
    ) -> Scalar:
        """
        Return the number of non-null values in the expression, only including values when the `where` condition is true.

        If `distinct` is True, then the number of unique values will be returned instead of the total count.

        If `approx` is True and `distinct` is True, method will use approximate count distinct function, which is faster but less accurate. This is only available for count distinct.
        """
        if distinct:
            if approx:
                return self._col.approx_nunique(where)
            else:
                return self._col.nunique(where)
        return self._col.count(where)

    def argmin(self, key: Value, where: BooleanValue | None = None) -> Scalar:
        """
        Return the value of key when the expression is at its minimum value.

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.argmin(where)

    def argmax(self, key: Value, where: BooleanValue | None = None) -> Scalar:
        """
        Return the value of key when the expression is at its maximum value.

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.argmax(where)

    def as_scalar(self) -> Scalar:
        """
        Convert the expression to a scalar value. Note that the execution of the scalar subquery will fail if the column expression contains more than one value.
        """
        return self._col.as_scalar()

    def first(self, where: BooleanValue | None = None) -> Scalar:
        """
        Return the first value in the expression.

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.first(where)

    def last(self, where: BooleanValue | None = None) -> Scalar:
        """
        Return the last value in the expression.

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.last(where)

    def lag(
        self, offset: int | IntegerValue | None = None, default: Value | None = None
    ) -> Scalar:
        """
        Return the row located at offset rows before the current row. If no row exists at offset, the default value is returned.
        """
        return self._col.lag(offset, default)

    def lead(
        self, offset: int | IntegerValue | None = None, default: Value | None = None
    ) -> Scalar:
        """
        Return the row located at offset rows after the current row. If no row exists at offset, the default value is returned.
        """
        return self._col.lead(offset, default)

    def max(self, where: BooleanValue | None = None) -> Scalar:
        """
        Return the maximum value of the expression

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.max(where)

    def min(self, where: BooleanValue | None = None) -> Scalar:
        """
        Return the minimum value of the expression

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.min(where)

    def mode(self, where: BooleanValue | None = None) -> Scalar:
        """
        Return the mode value of the expression

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.mode(where)

    def nth(self, n: int | IntegerValue) -> Scalar:
        """
        Return the nth value of the expression
        """
        return self._col.nth(n)

    def quantile(
        self,
        quantiles: float | NumericValue | list[NumericValue | float],
        where: BooleanValue | None = None,
    ) -> Scalar:
        """
        Return value at the given quantile. If multiple quantiles are specified, then the output will be an array of values.

        The output of this method a discrete quantile if the input is an float, otherwise it is a continuous quantile.
        """
        return self._col.quantile(quantiles, where)

    def like(
        self,
        patterns: str | StringValue | Iterable[str | StringValue],  # type: ignore[valid-type]
        case_sensitive: bool = True,
    ) -> BooleanValue:
        """
        This function is modeled after SQL's `LIKE` and `ILIKE` directives. Use `%` as a
        multiple-character wildcard or `_` as a single-character wildcard.

        For regular expressions, use `re.search`.
        """
        if case_sensitive:
            return self._col.like(patterns)
        return self._col.ilike(patterns)

    def combine(self, sep: str = ", ", where=None) -> Value:  # type: ignore[valid-type]
        """
        Combine the expression into a single string using the specified separator.

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.group_concat(sep, where)

    def collect(self, where: BooleanValue | None = None) -> ArrayColumn:
        """
        Collect the expression into an array.

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.collect()

    def sum(self, where: BooleanValue | None = None) -> VinylColumn:
        """
        Return the sum of the expression

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.sum(where)

    def mean(self, where: BooleanValue | None = None) -> VinylColumn:
        """
        Return the mean of the expression

        If a `where` condition is specified, method only considers rows meeting the `where` condition.
        """
        return self._col.mean(where)

    # inner classes
    @_transform_all_methods
    class MathFunctions:
        _col: NumericValue

        def __init__(self, col: Column):
            self._col = col

        def fabs(self) -> VinylColumn:
            """
            Return the absolute value of the expression.
            """
            return self._col.abs()

        def acos(self) -> VinylColumn:
            """
            Return the arc cosine of the expression.
            """
            return self._col.acos()

        def asin(self) -> VinylColumn:
            """
            Return the arc sine of the expression.
            """
            return self._col.asin()

        def atan(self) -> VinylColumn:
            """
            Return the arc tangent of the expression.
            """
            return self._col.atan()

        def atan2(self, other: NumericValue) -> VinylColumn:
            """
            Return the two-argument arc tangent of the expression.
            """
            return self._col.atan2(other)

        def ceil(self) -> VinylColumn:
            """
            Return the smallest integer value not less than the expression.
            """
            return self._col.ceil()

        def cos(self) -> VinylColumn:
            """
            Return the cosine of the expression.
            """
            return self._col.cos()

        def cot(self) -> VinylColumn:
            """
            Return the cotangent of the expression.
            """
            return self._col.cot()

        def degrees(self) -> VinylColumn:
            """
            Convert radians to degrees.
            """
            return self._col.degrees()

        def exp(self) -> VinylColumn:
            """
            Return the expression raised to the power of e.
            """
            return self._col.exp()

        def floor(self) -> VinylColumn:
            """
            Return the largest integer value not greater than the expression.
            """
            return self._col.floor()

        def log(self, base: NumericValue | None = None) -> VinylColumn:
            """
            Return the logarithm of the expression. If base is specified, then the logarithm will be taken to that base. Otherwise, the natural logarithm is taken.
            """
            return self._col.log(base)

        def log10(self) -> VinylColumn:
            """
            Return the base-10 logarithm of the expression.
            """
            return self._col.log10()

        def log2(self) -> VinylColumn:
            """
            Return the base-2 logarithm of the expression.
            """
            return self._col.log2()

        def radians(self) -> VinylColumn:
            """
            Convert degrees to radians.
            """
            return self._col.radians()

        def round(self, digits: int | IntegerValue | None = None) -> VinylColumn:
            """
            Round the expression to the specified number of decimal places. If digits is not specified, then the expression is rounded to the nearest integer.
            """
            return self._col.round(digits)

        def sign(self) -> VinylColumn:
            """
            Return the sign of the expression.
            """
            return self._col.sign()

        def sin(self) -> VinylColumn:
            """
            Return the sine of the expression.
            """
            return self._col.sin()

        def sqrt(self) -> VinylColumn:
            """
            Return the square root of the expression.
            """
            return self._col.sqrt()

        def tan(self) -> VinylColumn:
            """
            Return the tangent of the expression.
            """
            return self._col.tan()

        # math library functions added

        def copysign(self, other: NumericValue) -> VinylColumn:
            """
            Return the expression with the sign of the other expression.
            """
            return abs(self._col) * self._col.sign(other)

        def isclose(
            self, other: NumericValue, rel_tol: float = 1e-09, abs_tol: float = 0.0
        ) -> VinylColumn:
            """
            Return True if the values are close to each other and False otherwise.
            """
            return abs(self._col - other) <= max(
                rel_tol * max(abs(self._col), abs(other)), abs_tol
            )

        def isqrt(self) -> VinylColumn:
            """
            Return the integer square root of the expression.
            """
            return self._col.sqrt().floor()

        def ldexp(self, other: NumericValue) -> VinylColumn:
            """
            Return the expression multiplied by 2 to the power of the other expression.
            """
            return self._col * (2**other)

        def modf(self) -> VinylColumn:
            """
            Return the fractional and integer parts of the expression.
            """

            return VinylColumn([self._col - self._col.floor(), self._col.floor()])

        def prod(self, *args: NumericValue) -> VinylColumn:
            """
            Return the product of the expression and the other expressions.
            """
            out = self._col
            for arg in args:
                out *= arg
            return out

        def remainder(self, other: int | IntegerValue) -> VinylColumn:
            """
            Return the remainder of the expression divided by the other expression.
            """
            return self._col % other

        def sumprod(self, other: NumericValue) -> VinylColumn:
            """
            Return the sum of the product of the expression and the other expression.
            """
            return (self._col * other).sum()

        def trunc(self) -> VinylColumn:
            """
            Return the truncated value of the expression.
            """
            return self._col.trunc()

        def cbrt(self) -> VinylColumn:
            """
            Return the cube root of the expression.
            """
            return self._col ** (1 / 3)

        def exp2(self) -> VinylColumn:
            """
            Return 2 raised to the power of the expression.
            """
            return 2**self._col

        def expm1(self) -> VinylColumn:
            """
            Return e raised to the power of the expression minus 1.
            """
            return self._col.exp() - 1

        def pow(self, other: NumericValue) -> VinylColumn:
            """
            Return the expression raised to the power of the other expression.
            """
            return self._col**other

        def norm(self, type_: Literal["L1", "L2"] = "L2") -> VinylColumn:
            """
            Return the L2 norm of the expression.
            """
            if type_ == "L2":
                return (self._col**2).sum().sqrt()
            elif type_ == "L1":
                return self._col.abs().sum()
            else:
                raise ValueError("Invalid norm type")

        def dist(
            self,
            other: NumericValue,
            type_: Literal["Euclidian", "Manhattan", "Cosine"],
        ) -> VinylColumn:
            """
            Return the distance between the expression and the other expression.
            """
            if type_ == "Euclidian":
                return (self - VinylColumn(other)).math.norm("L2")

            if type_ == "Manhattan":
                return (self._col - VinylColumn(other)).math.norm("L1")

            if type_ == "Cosine":
                return (self._col * other).sum() / (
                    self.norm("L2") * VinylColumn(other).math.norm("L2")
                )

        def acosh(self) -> VinylColumn:
            """
            Return the inverse hyperbolic cosine of the expression.
            """
            return (self._col + (self._col**2 - 1).sqrt()).log()

        def asinh(self) -> VinylColumn:
            """
            Return the inverse hyperbolic sine of the expression.
            """
            return (self._col + (self._col**2 + 1).sqrt()).log()

        def atanh(self) -> VinylColumn:
            """
            Return the inverse hyperbolic tangent of the expression.
            """
            return ((1 + self._col) / (1 - self._col)).log() / 2

        def cosh(self) -> VinylColumn:
            """
            Return the hyperbolic cosine of the expression.
            """
            return (self._col.exp() + (-self._col).exp()) / 2

        def sinh(self) -> VinylColumn:
            """
            Return the hyperbolic sine of the expression.
            """
            return (self._col.exp() - (-self._col).exp()) / 2

        def tanh(self) -> VinylColumn:
            """
            Return the hyperbolic tangent of the expression.
            """
            return self.sinh() / self.cosh()

    @_transform_all_methods
    class URLFunctions:
        _col: StringValue

        def __init__(self, col: VinylColumn):
            self._col = col

        def authority(self) -> VinylColumn:
            """
            Return the authority of the expression.
            """
            return self._col.authority()

        def file(self) -> VinylColumn:
            """
            Parse a URL and extract the file.
            """
            return self._col.file()

        def fragment(self) -> VinylColumn:
            """
            Parse a URL and extract fragment identifier.
            """
            return self._col.fragment()

        def host(self) -> VinylColumn:
            """
            Parse a URL and extract the host.
            """
            return self._col.host()

        def path(self) -> VinylColumn:
            """
            Parse a URL and extract the path.
            """
            return self._col.path()

        def protocol(self) -> VinylColumn:
            """
            Parse a URL and extract the protocol.
            """
            return self._col.protocol()

        def query(self, key: str | StringValue | None = None) -> VinylColumn:
            """
            Parse a URL and extract the query string. If a key is specified, then the value of that key is returned. Otherwise, the entire query string is returned.
            """
            return self._col.query(key)

        def userinfo(self) -> VinylColumn:
            """
            Parse a URL and extract the userinfo.
            """
            return self._col.userinfo()

    @_transform_all_methods
    class RegexFunctions:
        _col: StringValue

        def __init__(self, col: VinylColumn):
            self._col = col

        def extract(
            self, pattern: str | StringValue, index: int | IntegerValue
        ) -> VinylColumn:
            """
            Return the specified match at index from a regex pattern.

            The behavior of this function follows the behavior of Python’s match objects: when index is zero and there’s a match, return the entire match, otherwise return the content of the index-th match group.
            """
            return self._col.re_extract(pattern, index)

        def replace(
            self, pattern: str | StringValue, replacement: str | StringValue
        ) -> VinylColumn:
            """
            Replace the matches of a regex pattern with a replacement string.
            """
            return self._col.re_replace(pattern, replacement)

        def search(self, pattern: str | StringValue) -> VinylColumn:
            """
            Returns True if the regex pattern matches a string and False otherwise.
            """
            return self._col.re_search(pattern)

        def split(self, pattern: str | StringValue) -> VinylColumn:
            """
            Split the expression using a regex pattern.
            """
            return self._col.re_split(pattern)

    @_transform_all_methods
    class StringFunctions:
        _col: StringValue

        def __init__(self, col: VinylColumn):
            self._col = col

        def ord(self) -> VinylColumn:
            """
            Return the unicode code point of the first character of the expression.
            """
            return self._col.ascii_str()

        def capitalize(self) -> VinylColumn:
            """
            Return a copy of the expression with the first character capitalized and the rest lowercased.
            """
            return self._col.capitalize()

        def contains(self, substr: str | StringValue) -> VinylColumn:
            """
            Return True if the expression contains the substring and False otherwise.
            """
            return self._col.contains(substr)

        def convert_base(
            self, from_base: IntegerValue, to_base: IntegerValue
        ) -> VinylColumn:
            """
            Convert the expression from one base to another.
            """
            return self._col.convert_base(from_base, to_base)

        def endswith(self, end: str | StringValue) -> VinylColumn:
            """
            Return True if the expression ends with the specified suffix and False otherwise.
            """
            return self._col.endswith(end)

        def find_in_set(self, str_list: list[str]) -> VinylColumn:
            """
            Return the position of the first occurrence of the expression in the list of strings.
            """
            return self._col.find_in_set()

        def find(
            self,
            substr: str | StringValue,
            start: int | IntegerValue | None = None,
            end: int | IntegerValue | None = None,
        ) -> VinylColumn:
            """
            Return the position of the first occurrence of substring. Search is limited to the specified start and end positions, if provided. All indexes are 0-based.
            """
            return self._col.find(substr, start, end)

        # standard library functions added

        def format(self, *args: Any, **kwargs: Any) -> VinylColumn:
            """
            Return a formatted string using the expression as a format string.

            Note that only a subset of the python format string syntax is supported.
            ```
            *Supported*
            - {0}, {1}, .. {n} for args replacements
            - {key} for kwargs replacements

            **Not Supported**
            - conversion flags (e.g. "Harold's a clever {0\!s}")
            - implicit references (e.g. "From {} to {}")
            - positional arguments / attributes (e.g. {0.weight} or {players[0]})
            ```
            """
            base: StringValue = self._col
            for i, arg in enumerate(args):
                base = base.replace(f"{{{i}}}", arg)

            for k, v in kwargs.items():
                base = base.replace(f"{{{k}}}", v)

            return base

        def to_strptime(self, format: str | StringValue) -> VinylColumn:
            """
            Parse a string into a timestamp using the specified strptime format.
            """
            return self._col.to_timestamp(format)

        def hash(
            self,
            algorithm: Literal["md5", "sha1", "sha256", "sha512"] = "sha256",
            return_type: Literal["bytes", "hex"] = "hex",
        ) -> VinylColumn:
            """
            Return the hash of the expression using the specified algorithm.
            """
            if return_type == "bytes":
                return self._col.hashbytes(algorithm)
            return self._col.hexdigest(algorithm)

        def join(self, strings: list[str | StringValue]) -> VinylColumn:
            """
            Concatenate the elements of the list using the provided separator.
            """
            return self._col.join(strings)

        def len(self, substr: str | StringValue) -> VinylColumn:
            """
            Return the number of non-overlapping occurrences of substring in the expression.
            """
            return self._col.length()

        def levenshtein(self, other: str | StringValue) -> VinylColumn:
            """
            Return the Levenshtein distance between the expression and the other string.
            """
            return self._col.levenshtein(other)

        def lower(self) -> VinylColumn:
            """
            Return a copy of the expression with all characters lowercased.
            """
            return self._col.lower()

        def ljust(
            self,
            length: int | IntegerValue,
            fillchar: str | StringValue | None = " ",
        ) -> VinylColumn:
            """
            Return the expression padded with the provided fill character to the specified length.
            """
            if fillchar is None:
                return self._col[:length]
            else:
                return self._col.lpad(length, fillchar)

        def lstrip(self) -> VinylColumn:
            """
            Return a copy of the expression with leading whitespace removed.

            Note: doesn't support removing specific characters like the standard library function.
            """
            return self._col.lstrip()

        def repeat(self, n: int | IntegerValue) -> VinylColumn:
            """
            Return the expression repeated `n` times.
            """
            return self._col.repeat(n)

        def replace(
            self, pattern: str | StringValue, replacement: str | StringValue
        ) -> VinylColumn:
            """
            Replace the matches of an exact (non-regex) pattern with a replacement string.
            """
            return self._col.replace(pattern, replacement)

        def rjust(
            self,
            length: int | IntegerValue,
            fillchar: str | StringValue | None = " ",
        ) -> VinylColumn:
            """
            Return the expression padded with the provided fill character to the specified length.
            """
            if fillchar is None:
                return self._col[-length:]
            else:
                return self._col.rpad(length, fillchar)

        def rstrip(self) -> VinylColumn:
            """
            Return a copy of the expression with trailing whitespace removed.

            Note: doesn't support removing specific characters like the standard library function.
            """
            return self._col.rstrip()

        def split(self, delimiter: str | StringValue) -> VinylColumn:
            """
            Split the expression using the specified delimiter.
            """
            return self._col.split(delimiter)

        def startswith(self, start: str | StringValue) -> VinylColumn:
            """
            Return True if the expression starts with the specified prefix and False otherwise.
            """
            return self._col.startswith(start)

        def strip(self) -> VinylColumn:
            """
            Return a copy of the expression with leading and trailing whitespace removed.
            """
            return self._col.strip()

        def reverse(self) -> VinylColumn:
            """
            Return a copy of the expression with the characters reversed.
            """
            return self._col.reverse()

        # python standard library additions
        def center(
            self, width: int | IntegerValue, fillchar: str | StringValue | None = None
        ) -> VinylColumn:
            """
            Return the expression centered in a string of length width. Padding is done using the specified fill character.
            """
            if fillchar is None:
                return self._col.substr((self._col.length() - width) // 2, width)
            adj: VinylColumn = self.ljust(width, fillchar).str.rjust(
                width, fillchar
            )  # guarantees this is long enough
            return adj._col.substr((adj._col.length() - width) // 2, width)

        def substr(
            self, start: int | IntegerValue, length: int | IntegerValue | None = None
        ) -> StringValue:
            """
            Return a substring of the expression starting at the specified index and optionally ending at the specified index.
            """
            return self._col.substr(start, length)

        def upper(self) -> VinylColumn:
            """
            Return a copy of the expression with all characters uppercased.
            """
            return self._col.upper()

    @_transform_all_methods
    class ArrayFunctions:
        _col: ArrayColumn

        def __init__(self, col: VinylColumn):
            self._col = col

        def unnest(self) -> VinylColumn:
            """
            Unnest the array into a new table.

            Note: Rows with empty arrays are dropped in the output.
            """
            return self._col.unnest()

        def join(self, sep: str) -> VinylColumn:
            """
            Concatenate the elements of the array using the provided separator.
            """
            return self._col.join(sep)

        def filter(
            self, predicate: Callable[[Value], bool] | BooleanValue
        ) -> VinylColumn:
            """
            Return a new array containing only the elements of the original array for which the predicate is true.
            """
            return self._col.filter(predicate)

        def flatten(self) -> VinylColumn:
            """
            Remove one level of nesting from the array.
            """
            return self._col.flatten()

        def index(self, value: Value) -> VinylColumn:
            """
            Return the position of the first occurrence of the value in the array.
            """
            return self._col.index(value)

        def len(self) -> VinylColumn:
            """
            Return the length of the array.
            """
            return self._col.length()

        def map(self, func: Callable[[Value], Value]) -> VinylColumn:
            """
            Apply the function to each element of the array.

            Note: also supports more complex callables like functools.partial and lambdas with closures
            """
            return self._col.map(func)

        def remove(self, value: Value) -> VinylColumn:
            """
            Return a new array with all occurrences of the value removed. Note that in the python standard library, this method only removes the first occurrence.
            """
            return self._col.remove(value)

        def repeat(self, n: int | IntegerValue) -> VinylColumn:
            """
            Return the array repeated `n` times.
            """
            return self._col.repeat(n)

        def sort(self) -> VinylColumn:
            """
            Return a new array with the elements sorted.
            """
            return self._col.sort()

        def union(self, other: ArrayColumn) -> VinylColumn:
            """
            Return a new array with the elements of both arrays, with duplicates removed.
            """
            return self._col.union(other)

        def unique(self) -> VinylColumn:
            """
            Return a new array with the duplicate elements removed.
            """
            return self._col.unique()

        def zip(self, other: ArrayValue, *others: ArrayValue) -> VinylColumn:
            """
            Return a new array with the elements of the original array and the other arrays zipped together.

            The combined map will have f1, f2, f3, etc. as the keys.
            """
            return self._col.zip(other, *others)

        # new method
        def del_(self, index: int | IntegerValue) -> VinylColumn:
            """
            Remove the element at the specified index from the array.
            """
            return self._col[:index] + self._col[index + 1 :]

        def insert(self, index: int | IntegerValue, value: Value) -> VinylColumn:
            """
            Insert the value at the specified index in the array.
            """
            return self._col[:index] + [value] + self._col[index:]

        def max(self) -> VinylColumn:
            """
            Return the maximum value of the array
            """
            return self._col.sort()[-1]

        def min(self) -> VinylColumn:
            """
            Return the minimum value of the array
            """
            return self._col.sort()[0]

    @_transform_all_methods
    class MapFunctions:
        _col: MapColumn

        def __init__(self, col: VinylColumn):
            self._col = col

        def contains(self, key: int | str | IntegerValue | StringValue) -> VinylColumn:
            """
            Return True if the map contains the specified key and False otherwise.
            """
            return self._col.contains(key)

        def get(self, key: Value, default: Value | None = None) -> VinylColumn:
            """
            Return the value of the specified key. If the key is not found, the default value is returned.
            """
            return self._col.get(key, default)

        def keys(self) -> VinylColumn:
            """
            Return the keys of the map.
            """
            return self._col.keys()

        def len(self) -> VinylColumn:
            """
            Return the length of the map.
            """
            return self._col.length()

        def values(self) -> VinylColumn:
            """
            Return the values of the map.
            """
            return self._col.values()

    @_transform_all_methods
    class StructFunctions:
        _col: StructColumn

        def __init__(self, col: VinylColumn):
            self._col = col

        @property
        def fields(self) -> VinylColumn:
            """
            Return a mapping from the field name to the field type of the struct
            """
            return self._col.fields

        @property
        def names(self) -> VinylColumn:
            """
            Return the names of the struct fields.
            """
            return self._col.names

        @property
        def types(self) -> VinylColumn:
            """
            Return the types of the struct fields.
            """
            return self._col.types

        def destructure(self) -> VinylColumn:
            """
            Destructure a StructValue into the corresponding struct fields.

            When assigned, a destruct value will be destructured and assigned to multiple columns.
            """
            return self._col.destructure()

        def lift(self) -> VinylColumn:
            """
            Project the fields of self into a table.

            This method is useful when analyzing data that has deeply nested structs or arrays of structs. lift can be chained to avoid repeating column names and table references.
            """
            return self._col.lift()

    @_transform_all_methods
    class TemporalFunctions:
        _col: TimestampValue | DateValue | IntervalValue

        def __init__(self, col: VinylColumn):
            self._col = col

        def extract(
            self,
            unit: Literal[
                "year",
                "quarter",
                "month",
                "week_of_year",
                "day",
                "day_of_year",
                "hour",
                "minute",
                "second",
                "microsecond",
                "millisecond",
            ],
        ) -> VinylColumn:
            """
            Extract the specified component from the datetime expression.
            """
            if unit == "year":
                return self._col.year()

            elif unit == "quarter":
                return self._col.quarter()

            elif unit == "month":
                return self._col.month()

            elif unit == "week_of_year":
                return self._col.week()

            elif unit == "day":
                return self._col.day()

            elif unit == "day_of_year":
                return self._col.day_of_year()

            elif unit == "hour":
                return self._col.hour()

            elif unit == "minute":
                return self._col.minute()

            elif unit == "second":
                return self._col.second()

            elif unit == "microsecond":
                return self._col.microsecond()

            elif unit == "millisecond":
                return self._col.millisecond()

            else:
                raise ValueError("Invalid unit")

        def floor(
            self,
            years: int | IntegerValue | None = None,
            quarters: int | IntegerValue | None = None,
            months: int | IntegerValue | None = None,
            weeks: int | IntegerValue | None = None,
            days: int | IntegerValue | None = None,
            hours: int | IntegerValue | None = None,
            minutes: int | IntegerValue | None = None,
            seconds: int | IntegerValue | None = None,
            milliseconds: int | IntegerValue | None = None,
            offset: IntervalValue | None = None,
        ) -> VinylColumn:
            """
            Round down the datetime expression to the specified unit. If an offset is specified, then the datetime will be rounded down to the nearest unit after the offset.

            If multiple units are specified, these will be combined together to form the unit. E.g. 1 quarter and 1 month will be transformed to 4 months.
            """
            unit_keys = ["Y", "Q", "M", "W", "D", "h", "m", "s", "ms", "us"]
            specified_units = [
                years,
                quarters,
                months,
                weeks,
                days,
                hours,
                minutes,
                seconds,
                milliseconds,
            ]
            if all(unit is None for unit in specified_units):
                raise ValueError("At least one unit must be specified")
            elif (
                specified_units.count(1) == 1
                and sum([u for u in specified_units if u is not None]) == 1
                and offset is None
            ):
                # can use truncate, which is supported by more backends
                specified_unit_key = [
                    unit_keys[i] for i, unit in enumerate(specified_units) if unit == 1
                ][0]
                return self._col.truncate(specified_unit_key)
            else:
                if not isinstance(self._col, TimestampValue) and not isinstance(
                    self._col, Deferred
                ):
                    raise ValueError(
                        f"Only simple units (e.g. 1 day, 1 week) are supported for {type(self._col)}. Please cast to a timestamp before using this method or use a simple unit instead."
                    )
                return self._col.bucket(
                    years=years,
                    quarters=quarters,
                    months=months,
                    weeks=weeks,
                    days=days,
                    hours=hours,
                    minutes=minutes,
                    seconds=seconds,
                    milliseconds=milliseconds,
                    offset=offset,
                )

        def epoch_seconds(self) -> VinylColumn:
            """
            Return the number of seconds since the Unix epoch.
            """
            return self._col.epoch_seconds()

        def strftime(self, format: str | StringValue) -> VinylColumn:  # type: ignore[valid-type]
            """
            Format string may depend on the backend, but we try to conform to ANSI strftime.
            """
            return self._col.strftime(format)


def _demote_arg(arg: Any) -> Value:
    if isinstance(arg, VinylColumn):
        return arg._col
    elif isinstance(arg, dict):
        return {_demote_arg(k): _demote_arg(v) for k, v in arg.items()}
    elif isinstance(arg, list):
        return [_demote_arg(v) for v in arg]
    elif isinstance(arg, tuple):
        return tuple(_demote_arg(v) for v in arg)
    elif isinstance(arg, set):
        return {_demote_arg(v) for v in arg}

    return arg


def _demote_args(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        new_args = []
        for i, arg in enumerate(args):
            new_args.append(_demote_arg(arg))
        new_kwargs = {}
        for k, v in kwargs.items():
            new_kwargs[k] = _demote_arg(v)

        return func(self, *new_args, **new_kwargs)

    return wrapper


def _demote_args_outside_class(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        new_args = []
        for i, arg in enumerate(args):
            new_args.append(_demote_arg(arg))
        new_kwargs = {}
        for k, v in kwargs.items():
            new_kwargs[k] = _demote_arg(v)

        return func(*new_args, **new_kwargs)

    return wrapper



from __future__ import annotations

import importlib
import inspect
import pkgutil
import re
from types import ModuleType
from typing import Any

import ibis.expr.operations as ops
import ibis.expr.types as ir
import networkx as nx
import rustworkx as rx
from ibis.expr.datatypes import DataType
from vinyl.lib.column import _demote_arg
from vinyl.lib.utils.graph import _rustworkx_to_networkx
from vinyl.lib.utils.obj import table_to_python_class


def get_caller_module(back: int) -> ModuleType | None:
    frame = inspect.currentframe()
    for _ in range(back + 1):
        if frame is None:
            raise ValueError("Unable to find the module path for the current project")
        frame = frame.f_back
    return inspect.getmodule(frame)


class Field:
    _relations: rx.PyDiGraph = rx.PyDiGraph()
    _relations_node_dict: dict[str, int] = {}
    _source_class_dict: dict[ir.Table, Any] = {}
    _source_adj_tbl_dict: dict[str, Any] = {}  # will be VinylTable
    name: str | None
    type: DataType | None
    description: str | None
    unique: bool
    primary_key: bool
    foreign_key: Field | tuple[str, str] | None
    pii: bool
    _parent_name: str | None
    _parent_ibis_table: object | None
    _parent_vinyl_table: object | None

    def __init__(
        self,
        name: str | None = None,
        type: DataType | None = None,
        description: str | None = None,
        primary_key: bool = False,
        foreign_key: Field | None = None,
        unique: bool = False,
        pii: bool = False,
        parent_name: str | None = None,
        parent_ibis_table: ops.UnboundTable | None = None,
        parent_vinyl_table: Any | None = None,
    ):
        self._parent_name = parent_name
        self._parent_ibis_table = parent_ibis_table
        self._parent_vinyl_table = parent_vinyl_table
        self.name = name
        self.type = type
        self.description = description
        self.primary_key = primary_key
        self.unique = unique
        self.pii = pii
        self.foreign_key = self.process_foreign_key(foreign_key)

    def process_foreign_key(self, foreign_key):
        if foreign_key is None:
            return None

        elif isinstance(foreign_key, tuple):
            if "." not in foreign_key[0]:
                caller_module = get_caller_module(2)
                if caller_module is not None:
                    module_name = caller_module.__name__
                else:
                    raise ValueError(
                        "Unable to find the module path for the current project"
                    )
                parent_module_name = ".".join(module_name.split(".")[:-1])
                parent_module = importlib.import_module(parent_module_name)
                child_modules = []
                for finder, name, ispkg in pkgutil.iter_modules(
                    parent_module.__path__, parent_module_name + "."
                ):
                    child_modules.append(name)

                relevant_names = [
                    name
                    for name in child_modules
                    if table_to_python_class(name.split(".")[-1]) == foreign_key[0]
                ]
                if len(relevant_names) > 1:
                    raise ValueError(
                        f"Multiple classes with the name {foreign_key[0]} found in the parent module. Please specify the full module path."
                    )
                elif len(relevant_names) == 0:
                    raise ValueError(
                        f"No class with the name {foreign_key[0]} found in the parent module."
                    )
                else:
                    adj_module_helper = relevant_names[0], foreign_key[0]

            else:
                split_ = foreign_key[0].split(".")
                adj_module_helper = ".".join(split_[:-1]), split_[-1]

            relevant_module = importlib.import_module(adj_module_helper[0])
            try:
                cls = getattr(relevant_module, adj_module_helper[1])
                foreign_key = getattr(cls, foreign_key[1])
            except AttributeError:
                # ignore fk in case of circular imports or invalid column names
                return None
        elif isinstance(foreign_key, str):
            # get field from str
            field_as_list = foreign_key.split(".")
            modname = ".".join(field_as_list[:-2])
            clsname = field_as_list[-2]
            colname = field_as_list[-1]
            mod = importlib.import_module(modname)
            try:
                cls = getattr(mod, clsname)
                foreign_key = getattr(cls, colname)

            except AttributeError:
                # ignore fk in case of circular imports or invalid column names
                return None

        return _demote_arg(
            foreign_key
        )  # necessary because unprocessed foreign key may be VinylColumn, not Ibis Column

    def _update_source_class(self):
        self._source_class_dict[self._parent_vinyl_table] = (
            self._parent_vinyl_table._source_class
            if hasattr(self._parent_vinyl_table, "_source_class")
            else None
        )

    def _store_source_adj_tbl(
        self, vinyl_tbl: Any
    ):  # will be VinylTable, but can't state due to circular import
        hash_ = str(hash(self._parent_ibis_table))
        if hash_ not in self._source_adj_tbl_dict:
            self._source_adj_tbl_dict[hash_] = vinyl_tbl

    def _store_relations(self):
        self_parent_table_vinyl = self._source_adj_tbl_dict[
            str(hash(self._parent_ibis_table))
        ]
        self._add_node(self_parent_table_vinyl)
        if self.foreign_key is not None:
            foreign_key_op = self.foreign_key.op()
            # note for below, foriegn key class has been imported, so self.foreign_key is an ibis expression at this point
            foreign_key_parent_table_ibis = foreign_key_op.find(ops.UnboundTable)[
                0
            ].to_expr()
            foreign_key_name = foreign_key_op.args[-1]
            foreign_key_parent_table_vinyl = self._source_adj_tbl_dict[
                str(hash(foreign_key_parent_table_ibis))
            ]

            self._add_edge(
                foreign_key_parent_table_vinyl,
                self_parent_table_vinyl,
                (foreign_key_name, self.name),
            )

            if self.unique:
                # relationship is bidirectional, so we need to add the edge in both directions
                self._add_edge(
                    self_parent_table_vinyl,
                    foreign_key_parent_table_vinyl,
                    (self.name, foreign_key_name),
                )

    def _add_node(self, node: ir.Expr):
        hash_ = str(hash(node))
        if hash_ not in self._relations_node_dict:
            self._relations_node_dict[hash_] = self._relations.add_node(node)

    def _add_edge(
        self,
        node1: ir.Expr,
        node2: ir.Expr,
        data: Any,
    ):
        for node in [node1, node2]:
            self._add_node(node)

        self._relations.add_edge(
            self._relations_node_dict[str(hash(node1))],
            self._relations_node_dict[str(hash(node2))],
            data,
        )

    def asdict(self):
        out = {}
        for k, v in self.__dict__.items():
            if v is not None and not k.startswith("_"):
                out[k] = v

        return out

    def _update(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def _export_relations_to_networkx(
        cls, shorten_name: bool = True, filter: list[str] | None = None
    ) -> nx.Graph | nx.DiGraph | nx.MultiGraph | nx.MultiDiGraph:
        G = _rustworkx_to_networkx(cls._relations)
        # make sure node object is passed as data rather than name to prevent issues with renderer
        nx.set_node_attributes(G, {node: node for node in G.nodes()}, "node")

        if shorten_name or filter:
            nx.relabel_nodes(
                G, {node: node.get_name() for node in G.nodes()}, copy=False
            )

        if not filter:
            return G
        else:
            all_nodes = list(G.nodes())
            nodes_in_subgraph: set[str] = set()
            for fil in filter:
                regex = re.compile(fil)
                nodes_to_add = set(node for node in all_nodes if regex.search(node))
                nodes_in_subgraph = nodes_in_subgraph.union(nodes_to_add)

            return G.subgraph(nodes_in_subgraph)

    @classmethod
    def get(cls, base_col: ir.Value):
        op = base_col.op()
        source_tbl = op.args[0].to_expr()
        source_vinyl_table = Field._source_adj_tbl_dict[str(hash(source_tbl))]
        source_vinyl_class = source_vinyl_table._source_class
        return source_vinyl_class._unique_name, getattr(source_vinyl_class, op.args[1])



import functools
import keyword
import re
import threading
from collections import OrderedDict
from typing import Any


def _is_iterable(obj: Any) -> bool:
    try:
        iter(obj)
        return True
    except TypeError:
        return False


def get_var_name(variable, local_vars):
    return [var_name for var_name, value in local_vars.items() if value is variable]


def is_valid_class_name(name: str) -> bool:
    # Check if the name is a Python keyword
    if keyword.iskeyword(name):
        return False

    # remove t names to avoid clash with type definitions
    if name == "t":
        return False

    # Check if the name follows the identifier rules
    if not name[0].isalpha() and name[0] != "_":
        return False
    if not all(char.isalnum() or char == "_" for char in name[1:]):
        return False

    return True


def to_valid_class_name(name: str, previous_replacements: list[str] = []) -> str:
    # Step 1: Ensure it starts with a letter or underscore. If not, prepend an underscore.
    if (
        not name[0].isalpha() or name[0] == "t"
    ):  # remove t names to avoid clash with type definitions
        name = "_" + name

    # Step 2: Replace invalid characters with underscores.
    name = re.sub(r"[^a-zA-Z0-9_]", "_", name)

    # Step 3: Avoid Python keywords by appending an underscore if needed.
    if keyword.iskeyword(name):
        name += "_"

    # Step 4: Ensure the result is not empty. If it is, default to a placeholder name.
    if not name:
        name = "Class_"

    # Step 5: Avoid duplicate names by appending a number if needed.
    if name in previous_replacements:
        i = 1
        while f"{name}{i}" in previous_replacements:
            i += 1
        name = f"{name}{i}"

    return name


def table_to_python_class(table_name) -> str:
    if not is_valid_class_name(table_name):
        table_name = to_valid_class_name(table_name)
    return "".join([word.capitalize() for word in table_name.split("_")])


def find_all_class_descendants(cls):
    descendants = []
    # Start with direct subclasses
    subclasses = cls.__subclasses__()
    descendants.extend(subclasses)
    # Recursively find subclasses of subclasses
    for subclass in subclasses:
        descendants.extend(find_all_class_descendants(subclass))
    return descendants


def threadsafe_lru_cache(maxsize=128, typed=False):
    def decorator(func):
        cache = OrderedDict()
        cache_lock = threading.Lock()

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            key = functools._make_key(args, kwargs, typed)

            with cache_lock:
                if key in cache:
                    # Move the accessed item to the end to maintain LRU order
                    cache.move_to_end(key)
                    return cache[key]

            result = func(*args, **kwargs)

            with cache_lock:
                if key not in cache:
                    if len(cache) >= maxsize:
                        # Remove the first item in the order (the least recently used)
                        cache.popitem(last=False)
                    cache[key] = result

            return result

        return wrapped

    return decorator
